#
# Originally Win32::FFI::BrowseForFolder version 0.02 by
#   Mike Blazer, blazer@mail.nevalink.ru
#   http://base.dux.ru/guest/fno/perl/
# (except that site is dead, so try web.archive.org's copy of the page and
# code) with a couple of changes to make it actually work
#

package BrowseForFolder;

use FFI;
use FFI::Library;
use Carp;

require Exporter;
@ISA = qw(Exporter);

$VERSION = "0.03";

@EXPORT = qw(
   CSIDL_DESKTOP
   CSIDL_PROGRAMS
   CSIDL_CONTROLS
   CSIDL_PRINTERS
   CSIDL_PERSONAL
   CSIDL_FAVORITES
   CSIDL_STARTUP
   CSIDL_RECENT
   CSIDL_SENDTO
   CSIDL_BITBUCKET
   CSIDL_STARTMENU
   CSIDL_DESKTOPDIRECTORY
   CSIDL_DRIVES
   CSIDL_NETWORK
   CSIDL_NETHOOD
   CSIDL_FONTS
   CSIDL_TEMPLATES
   CSIDL_COMMON_STARTMENU
   CSIDL_COMMON_PROGRAMS
   CSIDL_COMMON_STARTUP
   CSIDL_COMMON_DESKTOPDIRECTORY
   CSIDL_APPDATA
   CSIDL_PRINTHOOD

   BIF_RETURNONLYFSDIRS
   BIF_DONTGOBELOWDOMAIN
   BIF_STATUSTEXT
   BIF_RETURNFSANCESTORS
   BIF_BROWSEFORCOMPUTER
   BIF_BROWSEFORPRINTER
   BIF_BROWSEINCLUDEFILES
);

$kernel32     ||= FFI::Library->new("kernel32") or die "kernel32: could not make an instance\n";
$shell32      ||= FFI::Library->new("shell32")  or die "shell32: could not make an instance\n";
$user32       ||= FFI::Library->new("user32")   or die "user32: could not make an instance\n";
$gdi32        ||= FFI::Library->new("gdi32")    or die "gdi32: could not make an instance\n";

sub new {
   my $class = 'BrowseForFolder';
   my %arg;

   if (@_ && UNIVERSAL::isa($_[0], "HASH")) {
	%arg = %{$_[0]};
   } else {
	%arg = @_;
   }

   my $self = bless {
	-initial  => (defined $arg{"-initial"}  ? $arg{"-initial"}  : 0),
	-folder   => (defined $arg{"-folder"}   ? $arg{"-folder"}   : undef),
	-flags    => (defined $arg{"-flags"}    ? $arg{"-flags"}    : 0),
	-regexp   => (defined $arg{"-regexp"}   ? $arg{"-regexp"}   : undef),
	-status   => (defined $arg{"-status"}   ? $arg{"-status"}   : undef),
	-title    => (defined $arg{"-title"}    ? $arg{"-title"}    : undef),
	-text     => (defined $arg{"-text"}     ? $arg{"-text"}     : undef),
	-width    => (defined $arg{"-width"}    ? $arg{"-width"}    : undef),
	-height   => (defined $arg{"-height"}   ? $arg{"-height"}   : undef),
	-callback => (defined $arg{"-callback"} ? $arg{"-callback"} : undef),
	-owner    => (defined $arg{"-owner"}    ? $arg{"-owner"}    : undef),
   }, $class;

   $self;
}

#=========================
sub CreateCallback {
   my $self = shift;

   if ($self->{"-callback"}) {
      UNIVERSAL::isa($self->{"-callback"}, "CODE") or
	 croak "RegisterCallback: parameter - not a code ref\n";
      return $self->{"-callback"};
   }

   $self->{"-folder"} ||
   $self->{"-title"}  ||
   $self->{"-status"} ||
   $self->{"-regexp"} ||
   $self->{"-width"}  ||
   $self->{"-height"} ||
   $self->{"-folder"} or return sub{0};

   sub {
    my ($hwnd, $uMsg, $lParam, $lpData) = @_;
    my $pszPath = "\0"x260;

    if ($uMsg == BFFM_INITIALIZED()) {

	 SetWindowText($hwnd, $self->{"-title"}) if defined $self->{"-title"};

	 if ($self->{"-width"} || $self->{"-height"}) {
	    my ($wl, $wt, $wr, $wb) = GetWindowRect($hwnd);
	    my ($cl, $ct, $cw, $ch) = GetClientRect($hwnd);
	    my ($ww, $wh) = ($wr - $wl, $wb - $wt);

	    my $tree = FindByClass($hwnd, 0, "SysTreeView32");
	    my ($tl, $tt, $tr, $tb) = GetWindowRect($tree);

	    my $x_border = $tl-$wl;

	    # text labels
	    my $label1 = FindByClass($hwnd, 0, "Static");
	    my $label2 = FindByClass($hwnd, $label1, "Static");

	    my ($l1l, $l1t, $l1r, $l1b) = GetWindowRect($label1);
	    my ($l2l, $l2t, $l2r, $l2b) = GetWindowRect($label2);
	    if ($l1t > $l2t) {
		 my @a = ($label1, $l1l, $l1t, $l1r, $l1b);
		 ($label1, $l1l, $l1t, $l1r, $l1b) = ($label2, $l2l, $l2t, $l2r, $l2b);
		 ($label2, $l2l, $l2t, $l2r, $l2b) = @a;
	    }
	    # buttons
	    my $but1 = FindByClass($hwnd, 0, "Button");
	    my $but2 = FindByClass($hwnd, $but1, "Button");

	    my ($b1l, $b1t, $b1r, $b1b) = GetWindowRect($but1);
	    my ($b2l, $b2t, $b2r, $b2b) = GetWindowRect($but2);
	    if ($b1l > $b2l) {
		 my @a = ($but1, $b1l, $b1t, $b1r, $b1b);
		 ($but1, $b1l, $b1t, $b1r, $b1b) = ($but2, $b2l, $b2t, $b2r, $b2b);
		 ($but2, $b2l, $b2t, $b2r, $b2b) = @a;
	    }

	    my $butt_both_w = $b2r - $b1l;
	    my $min_width   = $butt_both_w + 2*$x_border;

	    # resize main window
	    my $Ww = $self->{"-width"} || $ww;
	    $Ww < $min_width and $Ww = $min_width;
	    my $Wh = $self->{"-height"} || $wh;
	    $Wh < 100 and $Wh = 100;

	    WindowResize($hwnd, $Ww, $Wh);

	    # move buttons (coords within the parent's client area)
	    WindowMove($but1, $Ww-$x_border-$butt_both_w-4, $Wh-($wh-$ch)-($wb-$b1t));
	    WindowMove($but2, $Ww-$x_border-$butt_both_w-4+$b2l-$b1l, $Wh-($wh-$ch)-($wb-$b1t));

	    my $vert = 4;
	    my $tree_w = $Ww-2*($x_border+2);
	    # vertical adjusting
	    if ($self->{"-text"}) {
	       my $hi = CalcTextSize($label1, $self->{"-text"}, $tree_w);

	       WindowMove($label1, $x_border-2, $vert);
	       WindowResize($label1, $tree_w, $hi); $vert += $hi+6;
	    } else {
	       WindowMove($label1, $x_border-2, $vert);
	       WindowResize($label1, $tree_w, 0);
	    }
	    if ($self->{"-status"}) {
	       my $hi = CalcTextSize($label2, $self->{"-status"}, $tree_w);
	       WindowMove($label2, $x_border-2, $vert);
	       WindowResize($label2, $tree_w, $hi); $vert += $hi+6;
	    } else {
	       WindowMove($label2, $x_border-2, $vert);
	       WindowResize($label2, $tree_w, 0);
	    }

	    # resize tree
	    WindowMove($tree, $x_border, $vert);
	    my $tree_h = $Wh-($wh-$ch)-$vert-($wb-$tb);
	    WindowResize($tree, $tree_w, $tree_h);
	 }

  $PostMessageP ||= $user32->function("PostMessageA", "sLIIIp");
  $PostMessageP->($hwnd, BFFM_SETSELECTIONA(), 1, $self->{"-folder"}) if defined $self->{"-folder"};

	 return 0;
    }
    return 0 if $uMsg != BFFM_SELCHANGED();

    $PostMessage ||= $user32->function("PostMessageA", "sLIIII");

    unless ($SHGetPathFromIDList->($lParam, $pszPath)) {
      $PostMessage->($hwnd, BFFM_ENABLEOK(), 1, 0);
    } else {
      $pszPath =~ s/\0.*$//s;
      $PostMessage->($hwnd, BFFM_ENABLEOK(), 1, (! defined($self->{"-regexp"}) || $pszPath =~ /$self->{"-regexp"}/ ? 1 : 0));
    }

    $self->status_handler($hwnd, $label2, $lParam, $pszPath) if $self->{"-status"};

    0;
  };
}

#=========================
sub status_handler {
  my ($self, $parent_hwnd, $label_hwnd, $pIDList, $path) = @_;

  $PostMessageP ||= $user32->function("PostMessageA", "sLIIIp");
  $PostMessageP->($parent_hwnd, BFFM_SETSTATUSTEXTA(), 0, defined $path ? $path : "");
}

#=========================
sub RegisterCallback {
   my $self = shift;
   UNIVERSAL::isa($self->{"-callback"}, "CODE") or
	 croak "RegisterCallback: parameter - not a code ref\n";

   FFI::callback("sIIIII", $self->{"-callback"});
}

#=======================
sub show {
   my $self = shift;
   my ($hwndOwner, $pidlRoot, $pszDisplayName,
       $lpszTitle, $nFolder, $ulFlags,
       $lpfn, $lParam, $iImage, $pszPath)
      =
      (defined($self->{"-owner"}) ? $self->{"-owner"} : GetConsoleHandle() || 0,
       "\0" x 260, "\0" x 260,
       $self->{"-text"} || "\0", $self->{"-initial"}, $self->{"-flags"} || 0x0000,
       "\0", 11, 0, "\0" x 260);

   $nFolder = CSIDL_DRIVES() unless defined $nFolder;
   $self->{"-status"} and $ulFlags |= BIF_STATUSTEXT();


   $SHGetSpecialFolderLocation ||= $shell32->function("SHGetSpecialFolderLocation", "sILIp");
   $SHGetSpecialFolderLocation->($hwndOwner, $nFolder, $pidlRoot)
	 and return undef;

#   $pidlRoot =~ s/\0.*$//s;
#   $pidlRoot = hex unpack 'H*', join'', reverse split//, $pidlRoot;
  $pidlRoot = substr($pidlRoot, 0, 4);

   $self->{"-callback"} = $self->CreateCallback;
   $BrowseCallbackProc  = $self->RegisterCallback;

   my $browseinfo = pack 'LLppILLI',
      ($hwndOwner, $pidlRoot, $pszDisplayName, $lpszTitle,
       $ulFlags, $BrowseCallbackProc->addr(), $lParam, $iImage);

   $SHGetPathFromIDList ||= $shell32->function("SHGetPathFromIDList", "sILp");
   $SHBrowseForFolder   ||= $shell32->function("SHBrowseForFolder", "sLp");

   my $bool = $SHGetPathFromIDList->(
               $SHBrowseForFolder->($browseinfo),
               $pszPath
              );

   $pszPath =~ s/\0.*$//s;
   $bool ? $pszPath : undef;
}

#=============
sub GetConsoleHandle () {
  my $oldtitle = GetConsoleTitle();

  my $newtitle = sprintf "PERL-%d-%d", Win32::GetTickCount(), $$;
  SetConsoleTitle($newtitle);
  milli_sleep(40);
  my $hwnd = FindWindowByTitle($newtitle);
  SetConsoleTitle($oldtitle);
  $hwnd;
}

#=============
sub GetConsoleTitle () {
  $GetConsoleTitle ||= $kernel32->function("GetConsoleTitleA", "sLpL");
  my $title = "\0" x 1024;
  $GetConsoleTitle->($title, 1024);
  $title;
}

#=============
sub SetConsoleTitle ($) {
  $SetConsoleTitle ||= $kernel32->function("SetConsoleTitleA", "sLp");
  $SetConsoleTitle->(shift);
}

#=============
sub FindWindowByTitle ($) {
  $FindWindow ||= $user32->function("FindWindowA", "sLLp");
  my $hwnd = $FindWindow->(0, shift);
}

#=============
sub milli_sleep ($) {
   $SleepEx ||= $kernel32->function("SleepEx", "svLI");
   $SleepEx->(shift, 0);
}

#=============
sub GetWindowText ($) {
# arg - hwnd
  $GetWindowText ||= $user32->function("GetWindowTextA", "sIIpI");

  my $buf = "\0"x1024;
  my $ret = $GetWindowText->(shift, $buf, 1024);
  $ret ? substr($buf, 0, $ret) : undef;
}

#=============
sub SetWindowText ($$) {
# arg - hwnd, text
  $SetWindowText ||= $user32->function("SetWindowTextA", "sIIp");

  $SetWindowText->(shift, shift);
}

#=============
sub WindowResize ($$$) {
# arg - hwnd, cx, cy
  $SetWindowPos ||= $user32->function("SetWindowPos", "sIIIIIIII");

  $SetWindowPos->(shift, 0,0,0, shift, shift,
	 SWP_NOZORDER() | SWP_NOOWNERZORDER() | SWP_NOMOVE());
}

#=============
sub WindowMove ($$$) {
# arg - hwnd, left, top
  $SetWindowPos ||= $user32->function("SetWindowPos", "sIIIIIIII");

  $SetWindowPos->(shift, 0, shift, shift,0,0,
	 SWP_NOZORDER() | SWP_NOOWNERZORDER() | SWP_NOSIZE());
}

#=============
sub GetWindowRect ($) {
# arg - hwnd
  $GetWindowRect ||= $user32->function("GetWindowRect", "sIIp");

  my $buf = "\0"x32;

  $GetWindowRect->(shift, $buf);

  my ($l, $t, $r, $b) = unpack "LLLL", $buf;
}

#=============
sub GetClientRect ($) {
# arg - hwnd
  $GetClientRect ||= $user32->function("GetClientRect", "sIIp");

  my $buf = "\0"x32;

  $GetClientRect->(shift, $buf);

  my ($l, $t, $r, $b) = unpack "LLLL", $buf;
}

#=============
sub FindWindowEx ($$) {
# arg - hwnd(parent),hwnd(0 or last enumerated child)
  $FindWindowEx ||= $user32->function("FindWindowExA", "sIIIII");

  my $childhwnd = $FindWindowEx->(shift, shift, 0,0) or return;
}

#=============
sub PrintChilds ($) {
# arg - hwnd(parent)
  my ($parent, $childhwnd, $class, $name) = (shift, 0);

  while (1) {
     $childhwnd = FindWindowEx($parent, $childhwnd);
     last unless $childhwnd;

     my ($l, $t, $r, $b) = GetWindowRect($childhwnd);
     print "$childhwnd: ".GetClassName($childhwnd).": $l, $t, $r, $b\n";
  }
}

#=============
sub GetClassName ($) {
# arg - hwnd
  $GetClassName ||= $user32->function("GetClassNameA", "sIIpI");

  my $buf = "\0"x1024;
  my $ret = $GetClassName->(shift, $buf, 1024);
  $ret ? substr($buf, 0, $ret) : undef;
}

#=============
sub FindByClass ($$$) {
# arg - hwnd(parent),hwnd(0 or last enumerated child), class-name
  $FindByClass ||= $user32->function("FindWindowExA", "sIIIpI");

  my $childhwnd = $FindByClass->(shift, shift, shift, 0) or return;
}

#=============
sub CalcTextSize ($$$) {
# arg - hwnd, text, window-width(pixels)
# returns - required height(pix)
  my ($hwnd, $text, $line_len) = @_;

  $text =~s/(\r?\n)(?=\r?\n)/$1 /g; # to preserve trailing \n\n\n\n\n
  my @text = split /\r?\n/, $text;

  $GetDC                ||= $user32->function("GetDC", "sII");
  $ReleaseDC            ||= $user32->function("ReleaseDC", "sIII");
  $GetTextExtentPoint32 ||= $gdi32->function("GetTextExtentPoint32A", "sIIpIp");
  $GetStockObject       ||= $gdi32->function("GetStockObject", "sII");
  $SelectObject         ||= $gdi32->function("SelectObject", "sIII");


  my $hdc = $GetDC->($hwnd);
  my $sum = 0;
  my $hfont = $GetStockObject->(17); # DEFAULT_GUI_FONT
  $SelectObject->($hdc, $hfont);

  for (@text) {
     $_ = " " unless $_; # to count blank lines also

     my $buf  = "\0"x16;
     my $bool = $GetTextExtentPoint32->($hdc, $_, length, $buf);

     unless ($bool) {
        $ReleaseDC->($hwnd, $hdc); return $sum;
     }

     my ($cx, $cy) = unpack "LL", $buf;
     my $l = int($cx / $line_len) + ($cx % $line_len ? 1 : 0);

     $sum += $l*$cy;
  }

  $ReleaseDC->($hwnd, $hdc);
  $sum;
}

sub AUTOLOAD {
    return if $AUTOLOAD =~ /::DESTROY$/;
    my $self = shift;

    if (ref $self) {			 # method call
         my ($method)= $AUTOLOAD =~ /^.*::([^:]+)$/; $method = "-$method";

	 if (exists $self->{$method}) {
	     $self->{$method} = shift if @_;
	     defined $self->{$method} ? $self->{$method} : undef;
	 } else {
	     croak "$AUTOLOAD: function not found\n";
	 }
    } else {
	 croak "$AUTOLOAD: function not found\n";
    }
}

sub CSIDL_DESKTOP                   () { 0x0000 }
sub CSIDL_PROGRAMS                  () { 0x0002 }
sub CSIDL_CONTROLS                  () { 0x0003 }
sub CSIDL_PRINTERS                  () { 0x0004 }
sub CSIDL_PERSONAL                  () { 0x0005 }
sub CSIDL_FAVORITES                 () { 0x0006 }
sub CSIDL_STARTUP                   () { 0x0007 }
sub CSIDL_RECENT                    () { 0x0008 }
sub CSIDL_SENDTO                    () { 0x0009 }
sub CSIDL_BITBUCKET                 () { 0x000a }
sub CSIDL_STARTMENU                 () { 0x000b }
sub CSIDL_DESKTOPDIRECTORY          () { 0x0010 }
sub CSIDL_DRIVES                    () { 0x0011 }
sub CSIDL_NETWORK                   () { 0x0012 }
sub CSIDL_NETHOOD                   () { 0x0013 }
sub CSIDL_FONTS                     () { 0x0014 }
sub CSIDL_TEMPLATES                 () { 0x0015 }
sub CSIDL_COMMON_STARTMENU          () { 0x0016 }
sub CSIDL_COMMON_PROGRAMS           () { 0x0017 }
sub CSIDL_COMMON_STARTUP            () { 0x0018 }
sub CSIDL_COMMON_DESKTOPDIRECTORY   () { 0x0019 }
sub CSIDL_APPDATA                   () { 0x001a }
sub CSIDL_PRINTHOOD                 () { 0x001b }

sub WM_USER () { 0x0400 }
sub BFFM_SETSTATUSTEXTA     () { WM_USER() + 100 }
sub BFFM_ENABLEOK           () { WM_USER() + 101 }
sub BFFM_SETSELECTIONA      () { WM_USER() + 102 }
sub BFFM_SETSELECTIONW      () { WM_USER() + 103 }
sub BFFM_SETSTATUSTEXTW     () { WM_USER() + 104 }
sub BFFM_INITIALIZED        () { 1 }
sub BFFM_SELCHANGED         () { 2 }

sub BIF_RETURNONLYFSDIRS   () { 0x0001 } # For finding a folder to start document searching
sub BIF_DONTGOBELOWDOMAIN  () { 0x0002 } # For starting the Find Computer
sub BIF_STATUSTEXT         () { 0x0004 } # Includes a status area in the dialog box.
sub BIF_RETURNFSANCESTORS  () { 0x0008 }
sub BIF_EDITBOX            () { 0x0010 }
sub BIF_VALIDATE           () { 0x0020 }
sub BIF_BROWSEFORCOMPUTER  () { 0x1000 } # Browsing for Computers.
sub BIF_BROWSEFORPRINTER   () { 0x2000 } # Browsing for Printers
sub BIF_BROWSEINCLUDEFILES () { 0x4000 } # Browsing for Everything


sub SWP_NOSIZE          () { 0x0001 }
sub SWP_NOMOVE          () { 0x0002 }
sub SWP_NOZORDER        () { 0x0004 }
sub SWP_NOREDRAW        () { 0x0008 }
sub SWP_NOACTIVATE      () { 0x0010 }
sub SWP_FRAMECHANGED    () { 0x0020 }  #/* The frame changed: send WM_NCCALCSIZE */
sub SWP_SHOWWINDOW      () { 0x0040 }
sub SWP_HIDEWINDOW      () { 0x0080 }
sub SWP_NOCOPYBITS      () { 0x0100 }
sub SWP_NOOWNERZORDER   () { 0x0200 }  #/* Don't do owner Z ordering */
sub SWP_NOSENDCHANGING  () { 0x0400 }  #/* Don't send WM_WINDOWPOSCHANGING */

sub SWP_DRAWFRAME       () { SWP_FRAMECHANGED() }
sub SWP_NOREPOSITION    () { SWP_NOOWNERZORDER() }

sub SWP_DEFERERASE      () { 0x2000 }
sub SWP_ASYNCWINDOWPOS  () { 0x4000 }
