#
#
#  DO TGA OUTPUT!
#
#

package Convert;

use strict;
use warnings;

use IO::Scalar;
use IO::File;
use HTML::Entities();
use File::Spec;
require File::Spec::Win32;

# use lib 'e:/resrip/dxtc/dxtc/blib/lib';
# use lib 'e:/resrip/dxtc/dxtc/blib/arch';
#use DXTC; # C/XS extension for fast(ish) DXTC compression

#use lib 'e:/resrip/cmodule/blib/lib';
#use lib 'e:/resrip/cmodule/blib/arch';
use resrip; # C extension for fast(ish) DXTC compression and palettising

use ThreeDS; # for 3ds file output

################################################

sub xmb_to_xml {
  my ($input_filename, $output_filename, $params) = @_;

  my $comp_in;
  if ($params->{in_fh}) {
    $comp_in = $params->{in_fh};
    binmode $comp_in unless ref $comp_in eq 'IO::Scalar';
  } else {
    $comp_in = IO::File->new("< $input_filename") or die "Error opening $input_filename for input: $!";
    binmode $comp_in;
  }

  my $out;
  if ($params->{out_fh}) {
    $out = $params->{out_fh};
    binmode $out unless ref $out eq 'IO::Scalar';
  } else {
    $out = IO::File->new("> $output_filename") or die "Error opening $output_filename for output: $!";
    binmode $out;
  }

  my $raw_data = join('', $comp_in->getlines);

  my ($uncompressed, $alpha_version);
  my $in;
  if (substr($raw_data, 0, 2) eq 'X1') { # alpha/beta-style XMBs
    $uncompressed = 1;
    $in = IO::Scalar->new(\$raw_data);
    if (substr($raw_data, 16, 2) eq 'XN') { # alpha-style XMBs
      $alpha_version = 1;
    } else {
      $alpha_version = 0;
    }
  } else {
    $uncompressed = 0;
    $alpha_version = 0;
    my $data = main::decompress($raw_data, 'l33t_zlib');
    $in = IO::Scalar->new(\$data);
  }

  $comp_in->close unless $params->{in_fh};

  my $buffer;

  $out->print(qq{<?xml version="1.0" encoding="UTF-8"?>\r\n\r\n});

  $in->read($buffer, 2); # file_head
  $buffer eq 'X1' or die "Invalid XMB file!";

  $in->read($buffer, 14); # data_length, root_head, root_numbers

  if (not $alpha_version) {

    my @elements;
    $in->read($buffer, 4); # elements_count
    for my $el_id (0 .. unpack(L => $buffer)-1) {
      $in->read($buffer, 4); # element_X_length
      $in->read($buffer, 2 * unpack L => $buffer); # element_X
      push @elements, utf16to8($buffer);
    }
  
    my @params;
    $in->read($buffer, 4); # params_count
    for my $pa_id (0 .. unpack(L => $buffer)-1) {
      $in->read($buffer, 4); # param_X_length
      $in->read($buffer, 2 * unpack L => $buffer); # param_X
      if (length $buffer) {
        push @params, utf16to8($buffer);
      } else {
        push @params, 'xmlns'; # (to make the third parameter in en.xmb/etc work)
      }
    }
  
    xmb_to_xml__read_elements({
      in => $in, out => $out, tabs => '',
      elements => \@elements, params => \@params
    });
  } else {
    xmb_to_xml__read_elements({
      alpha_version => 1,
      in => $in, out => $out, tabs => ''
    });

  }

  $out->close unless $params->{out_fh};
  $in->close;
}

sub xmb_to_xml__read_elements {
  my $params = shift;
  my ($in, $out) = ($params->{in}, $params->{out});

  my $buffer;
  $in->read($buffer, 6); # element_X_head, element_X_length

  $in->read($buffer, 4); # element_X_content_length
  $in->read($buffer, 2 * unpack L => $buffer); # element_X_content
  my $content = xml_encode(utf16to8($buffer), '<>');

  my $name;
  if ($params->{alpha_version}) {
    $in->read($buffer, 4);
    $in->read($buffer, 2 * unpack L => $buffer);
    $name = utf16to8($buffer);
    $in->read($buffer, 4);
  } else {
    $in->read($buffer, 4); # element_X_name
    $name = $params->{elements}[unpack L => $buffer];
  }

  my @params;
  $in->read($buffer, 4); # element_X_params_count
  for my $pa_id (0 .. unpack(L => $buffer)-1) {
    my $name;
    if ($params->{alpha_version}) {
      $in->read($buffer, 4);
      $in->read($buffer, 2 * unpack L => $buffer);
      $name = utf16to8($buffer);
    } else {
      $in->read($buffer, 4); # element_X_param_X_name
      $name = $params->{params}[unpack L => $buffer];
    }
    $in->read($buffer, 4); # element_X_param_X_length
    $in->read($buffer, 2 * unpack L => $buffer); # element_X_param_X
    push @params, [ $name, xml_encode(utf16to8($buffer), '<>"') ];
  }

  $out->print("$params->{tabs}<$name",
              (map qq{ $_->[0]="$_->[1]"}, @params),
              ">$content");

  $in->read($buffer, 4); # element_X_subel_count
  my $sub_elements = unpack(L => $buffer);
  if ($sub_elements) {
    $out->print("\r\n");
    {
      my $old_tabs = $params->{tabs};
      $params->{tabs} .= "\t";
      for (0 .. $sub_elements-1) {
        xmb_to_xml__read_elements($params);
      }
      $params->{tabs} = $old_tabs;
    }
    $out->print(qq{$params->{tabs}</$name>\r\n});
  } else {
    $out->print(qq{</$name>\r\n});
  }
}


################################################

sub xml_to_xmb {
  my ($input_filename, $output_filename, $params) = @_;

  my $in;
  if ($params->{in_fh}) {
    $in = $params->{in_fh};
  } else {
    $in = IO::File->new("< $input_filename") or die "Error opening $input_filename for input: $!";
  }

  my $comp_out;
  if ($params->{out_fh}) {
    $comp_out = $params->{out_fh};
    binmode $comp_out unless ref $comp_out eq 'IO::Scalar';
  } else {
    $comp_out = IO::File->new("> $output_filename") or die "Error opening $output_filename for output: $!";
    binmode $comp_out;
  }

  require XML::Parser;

  my $parser = XML::Parser->new(Style => 'Tree', ErrorContext => 2);
  my $xml_tree;
  eval { $xml_tree = $parser->parse($in) };
  die qq{Error in XML code:$@\n} if $@;

  my (%params, %elements);
  my $xml_data = xml_to_xmb__preprocess(\%params, \%elements, $xml_tree);
  my @param_names = sort keys %params;
  my @element_names = sort keys %elements;

  $params{$param_names[$_]} = $_ for 0..$#param_names;
  $elements{$element_names[$_]} = $_ for 0..$#element_names;

  my $output;
  $output .= 'XR' . pack(LL => 4, 7);

  $output .= pack L => scalar @element_names;
  foreach my $element_name (map utf8to16($_), @element_names) {
    $output .= pack L => length($element_name) / 2;
    $output .= $element_name;
  }
  $output .= pack L => scalar @param_names;
  foreach my $param_name (map utf8to16($_), @param_names) {
    if ($param_name eq 'xmlns') {
      $output .= pack L => 0;
    } else {
      $output .= pack L => length($param_name) / 2;
      $output .= $param_name;
    }
  }
  $output .= xml_to_xmb__write(\%params, \%elements, $xml_data);

  $output = 'X1' . pack(L => length $output) . $output;

  $comp_out->print(main::compress($output, 'l33t_zlib'));

  $comp_out->close unless $params->{out_fh};
  $in->close unless $params->{in_fh};
}

sub xml_to_xmb__preprocess {
  my ($params, $elements, $node) = @_;
  my ($el_name, $el_contents) = @$node;
  $elements->{$el_name} = 1;
  my %attr_data = %{shift @$el_contents};
  $params->{$_} = 1 foreach keys %attr_data;

  $attr_data{$_} = utf8to16($attr_data{$_}) foreach keys %attr_data;

  my $content_text = '';
  my @subnodes;
  while (@$el_contents and my ($tag, $data) = (shift @$el_contents, shift @$el_contents)) {
    next if $tag eq '0' and $data =~ /^\s*$/;
    if ($tag eq '0') {
      $content_text .= utf8to16($data);
    } else {
      push @subnodes, xml_to_xmb__preprocess($params, $elements, [$tag, $data]);
    }
  }
  return { name => $el_name, params => \%attr_data, text => $content_text, subnodes => \@subnodes };
}

sub xml_to_xmb__write {
  my ($params, $elements, $node) = @_;

  my $return;
  $return .= pack L => length($node->{text}) / 2;
  $return .= $node->{text};
  $return .= pack L => $elements->{$node->{name}};

  $return .= pack L => scalar keys %{$node->{params}};
  foreach my $param_key (sort keys %{$node->{params}}) {
    $return .= pack LL => $params->{$param_key}, length($node->{params}{$param_key}) / 2;
    $return .= $node->{params}{$param_key};
  }

  $return .= pack L => scalar @{$node->{subnodes}};
  foreach my $subnode (@{$node->{subnodes}}) {
    $return .= xml_to_xmb__write($params, $elements, $subnode);
  }

  return 'XN' . pack(L => length($return)) . $return;
}

################################################

sub ddt_to_image {
  my ($input_filename, $output_filename, $params) = @_;

  my $in;
  if ($params->{in_fh}) {
    $in = $params->{in_fh};
    binmode $in unless ref $in eq 'IO::Scalar';
  } else {
    $in = IO::File->new("< $input_filename") or die "Error opening $input_filename for input: $!";
    binmode $in;
  }

  my $image_format;
  if ($output_filename) {
    $output_filename =~ /^.*\.(\w+)$/;
    $image_format = lc $1;
  } else {
    $image_format = $params->{imageformat} || 'bmp';
  }

  my $buffer;

  my ($image_data, $alpha_data);

  $in->read($buffer, 4); # head_text
  $buffer eq 'RTS3' or die "Invalid DDT file!";

  $in->read($buffer, 3); # pixel_format01/02/03
  my @pixel_formats = unpack C3 => $buffer;

  ($pixel_formats[2] == 3 and
    ($pixel_formats[1] == 0 or
     $pixel_formats[1] == 1 or
     $pixel_formats[1] == 4)) or
  ($pixel_formats[2] == 4 and
     $pixel_formats[1] == 0) or
  ($pixel_formats[2] == 5 and
     $pixel_formats[1] == 1) or
  ($pixel_formats[2] == 6 and
     $pixel_formats[1] == 4) or
  ($pixel_formats[2] == 7 and
     $pixel_formats[1] == 8) or
    die "This type of DDT (@pixel_formats) shouldn't exist!\n";

  $in->read($buffer, 1); # num_levels
  my $num_levels = unpack C => $buffer;

  $in->read($buffer, 8); # image_width, image_height
  my ($width, $height) = unpack LL => $buffer;

  my ($num_colours, $palette32_offset, $palette16_offset, $palette15_offset, $palette15b_offset, $palette12_offset);
  if ($pixel_formats[2] == 3) {
    $in->read($buffer, 4); # num_colours
    $num_colours = unpack L => $buffer;
  
    $in->read($buffer, 4); # palette32_offset
    $palette32_offset = unpack L => $buffer;
  
    $in->read($buffer, 4); # palette16_offset
    $palette16_offset = unpack L => $buffer;
  
    $in->read($buffer, 4); # palette15_offset
    $palette15_offset = unpack L => $buffer;
  
    $in->read($buffer, 4); # palette15b_offset
    $palette15b_offset = unpack L => $buffer;

    $in->read($buffer, 4); # palette12_offset
    $palette12_offset = unpack L => $buffer;
  }

  my @images;
  for my $image (0..$num_levels-1) {
    $in->read($buffer, 8); # image_%I_offset, image_%I_length
    push @images, [ unpack LL => $buffer ];
  }

  # Read the palette (when necessary)
  my @palette;

  if ($palette32_offset) {
    $in->seek($palette32_offset, 0);
    $in->read($buffer, $num_colours * 4);
    @palette = map main::colour_32_to_bytes_rev($_), unpack 'L*' => $buffer;
  } elsif ($pixel_formats[2] == 3 and $pixel_formats[1] == 0) {
    $in->seek($palette16_offset, 0);
    $in->read($buffer, $num_colours * 2);
    @palette = map main::colour_16_to_bytes_rev($_), unpack 'S*' => $buffer;
  } elsif ($pixel_formats[2] == 3 and $pixel_formats[1] == 1) {
    $in->seek($palette15b_offset, 0);
    $in->read($buffer, $num_colours * 2);
    @palette = map main::colour_15_to_bytes_rev($_), unpack 'S*' => $buffer;
  } elsif ($pixel_formats[2] == 3 and $pixel_formats[1] == 4) {
    $in->seek($palette12_offset, 0);
    $in->read($buffer, $num_colours * 2);
    @palette = map main::colour_12_to_bytes_rev($_), unpack 'S*' => $buffer;
  }

#  for my $image (0..$num_levels-1) {
  for my $image (0) { # only read the first (largest) image
    $in->seek($images[$image][0], 0);
    $in->read($buffer, $images[$image][1]);

    if ($pixel_formats[2] == 3 and $pixel_formats[1] == 4) {
      my $c = 0; # alpha-alpha-colour-colour-colour-colour bytes
      $alpha_data = join '', map chr($_), map +($_&240, ($_&15)<<4), grep !(int($c++/2)%3), unpack 'C*' => $buffer;
      $image_data = join '', map $palette[$_], grep int($c++/2)%3, unpack 'C*' => $buffer;

    } elsif ($pixel_formats[2] == 3) {
      $image_data = join '', map $palette[$_], unpack 'C*' => $buffer;

    } elsif ($pixel_formats[2] == 7) {
      $image_data = $buffer;
      $image_data =~ s/(.)/$1$1$1/sg; # triple everything

    } elsif ($pixel_formats[2] == 4 or $pixel_formats[2] == 5 or $pixel_formats[2] == 6) {
  
      $image_data = '';

      my (@colour_pixel, @colour_pixels, $data_byte);

      if ($pixel_formats[2] == 4) {

        for my $y (0..$height/4-1) {
          @colour_pixels = ();
          for my $x (0..$width/4-1) {
            @colour_pixel = map main::colour_16_to_bytes_rev($_), unpack S2 => substr($buffer, 8*($y*$width/4 + $x), 4);
            push @colour_pixel, pack C3 => map int((vec($colour_pixel[0], $_, 8)*2+vec($colour_pixel[1], $_, 8))/3), 0..2;
            push @colour_pixel, pack C3 => map int((vec($colour_pixel[0], $_, 8)+2*vec($colour_pixel[1], $_, 8))/3), 0..2;
            push @colour_pixels, @colour_pixel;
          }
          for my $sub_y (0..3) {
            for my $x (0..$width/4-1) {
              $data_byte = substr($buffer, 4+$sub_y+8*($y*$width/4 + $x), 1);
              $image_data .= $colour_pixels[$x*4+vec($data_byte, $_, 2)] for 0..3;
            }
          }
        }

      } elsif ($pixel_formats[2] == 5) {

        my $alpha_byte; # 1-bit alpha
        for my $y (0..$height/4-1) {
          for my $sub_y (0..3) {
            for my $x (0..$width/4-1) {
              $alpha_byte = substr($buffer, 4+10*(int($y*$width/4)+$x), 2);
              $alpha_data .= chr(255*vec($alpha_byte, $_+4*$sub_y, 1)) for 0..3;
            }
          }
        }

        for my $y (0..$height/4-1) {
          @colour_pixels = ();
          for my $x (0..$width/4-1) {
            @colour_pixel = map main::colour_15_to_bytes_rev($_), unpack S2 => substr($buffer, 10*($y*$width/4 + $x), 4);
            push @colour_pixel, pack C3 => map int((vec($colour_pixel[0], $_, 8)*2+vec($colour_pixel[1], $_, 8))/3), 0..2;
            push @colour_pixel, pack C3 => map int((vec($colour_pixel[0], $_, 8)+2*vec($colour_pixel[1], $_, 8))/3), 0..2;
            push @colour_pixels, @colour_pixel;
          }
          for my $sub_y (0..3) {
            for my $x (0..$width/4-1) {
              $data_byte = substr($buffer, 6+$sub_y+10*($y*$width/4 + $x), 1);
              $image_data .= $colour_pixels[$x*4+vec($data_byte, $_, 2)] for 0..3;
            }
          }
        }
  
      } elsif ($pixel_formats[2] == 6) {

        my $alpha_byte; # 4-bit alpha
        for my $y (0..$height/4-1) {
          for my $sub_y (0..3) {
            for my $x (0..$width/4-1) {
              $alpha_byte = substr($buffer, 4+16*(int($y*$width/4)+$x), 12);
              $alpha_data .= chr(vec($alpha_byte, $_+6*$sub_y, 4) << 4) for 0..3;
            }
          }
        }

        for my $y (0..$height/4-1) {
          @colour_pixels = ();
          for my $x (0..$width/4-1) {
            @colour_pixel = map main::colour_12_to_bytes_rev($_), unpack S2 => substr($buffer, 16*($y*$width/4 + $x), 4);
            push @colour_pixel, pack C3 => map int((vec($colour_pixel[0], $_, 8)*2+vec($colour_pixel[1], $_, 8))/3), 0..2;
            push @colour_pixel, pack C3 => map int((vec($colour_pixel[0], $_, 8)+2*vec($colour_pixel[1], $_, 8))/3), 0..2;
            push @colour_pixels, @colour_pixel;
          }
          for my $sub_y (0..3) {
            for my $x (0..$width/4-1) {
              $data_byte = substr($buffer, 6+3*$sub_y+16*($y*$width/4 + $x), 1);
              $image_data .= $colour_pixels[$x*4+vec($data_byte, $_, 2)] for 0..3;
            }
          }
        }
      }

    } else {
      die "Invalid pixel_format '@pixel_formats'!";
    }
  }

  $in->close unless $params->{in_fh};

  my $out;
  if ($params->{out_fh}) {
    $out = $params->{out_fh};
    binmode $out unless ref $out eq 'IO::Scalar';
  } else {
    $out = IO::File->new("> $output_filename") or die "Error opening $output_filename for output: $!";
    binmode $out;
  }

  my $output;
  if ($image_format eq 'bmp') {
    if (defined $alpha_data) {
      $alpha_data =~ s/(.)/$1$1$1/sg; # triple everything
      $image_data = $alpha_data . $image_data;
      $height *= 2;
    }
    $output = pack(C4LSL4SLSL5 => 0x42, 0x4d, 0x66, 0xb4,
                                  0, 0, 0x36, 0x28,
                                  $width, $height, 0x01, 0x18, 0,
                                  ($width*3 + $width%4)*$height, 0, 0, 0, 0);
    for my $y (0..$height-1) {
      $output .= substr($image_data, $y*$width*3, $width*3);
      $output .= "\0" x ($width % 4);
    }

  } elsif ($image_format eq 'tga') {

    if (defined $alpha_data) {
      $output = pack(S8C2 => 0,2,0,0,0,0, $width, $height, 32, 8);
      for my $y (0..$height-1) {
        $output .= substr($image_data, ($y*$width+$_)*3, 3) . substr($alpha_data, $y*$width+$_, 1)
          for 0..$width-1;
      }
    } else {
      $output = pack(S8C2 => 0,2,0,0,0,0, $width, $height, 24, 0);
      for my $y (0..$height-1) {
        $output .= substr($image_data, $y*$width*3, $width*3);
      }
    }

  } else {
    die qq{Unrecognised output image format "$image_format"!};
  }
  $out->print($output);

  $out->close unless $params->{out_fh};

  return { type => join('', @pixel_formats), levels => $num_levels };
}


################################################

sub image_to_ddt {
  my ($input_filename, $output_filename, $params) = @_;

  my $in;
  if ($params->{in_fh}) {
    $in = $params->{in_fh};
    binmode $in unless ref $in eq 'IO::Scalar';
  } else {
    $in = IO::File->new("< $input_filename") or die "Error opening $input_filename for input: $!";
    binmode $in;
  }

  my $image_format;
  if ($input_filename) {
    $input_filename =~ /^.*\.(\w+)$/;
    $image_format = lc $1;
  } else {
    $image_format = $params->{imageformat} || 'bmp';
  }

  # need to ask what format to use:
  # 0 0 3 -> terrain, usually 2 levels but 5 for water\d{4}
  # 2 0 3 -> various icons and ui, always 1 level
  # 3 0 3 -> a few ui, always 1 level
  # 4 0 3 -> terrain/waterbump\d{4}?, always 1 level
  # 2 1 3 -> various ui, icons, cursors, always 1 level
  # 2 4 3 -> icons, ui, always 1 level
  # 0 0 4 -> 1 or 4 or 5(usually) levels
  # 1 0 4 -> similar, though always 5
  # 2 0 4 -> ditto
  # 8 0 4 -> sfx shiny maps, always 5
  # 0 1 5 -> misc images, usually 1 or 5 or 4(once) levels
  # 1 1 5 -> ditto
  # 2 1 5 -> ditto, always 5
  # 3 1 5 -> ditto, always 5
  # 0 4 6 -> vaguely sprites, shadows, etc, 1 or 3 or 4 or 5
  # 1 4 6 -> ditto
  # 2 4 6 -> shadows, 1 level except for chatlogo(5)
  # 0 8 7 -> terrain/blend*, 2 levels

  my $got_alpha = 0;
  $got_alpha = 1 if $params->{format} =~ /..[56]/ or $params->{format} =~ /.43/;

  my ($width, $height, $raw_data, $alpha_data);
  if ($image_format eq 'bmp') {
    my $buffer;
    $in->seek(18, 0);
    $in->read($buffer, 8);
    ($width, $height) = unpack L2 => $buffer;

    my $offset = 54;

    if ($got_alpha) {
      $height /= 2;
      for my $y (0..$height-1) {
        $in->seek($offset + $y*$width*3 + $y*($width%4), 0);
        $in->read($buffer, $width*3);
        $alpha_data .= substr($buffer, $_*3, 1) for 0..$width-1;
      }
      $offset += $height*$width*3 + $height*($width%4);
     }

    for my $y (0..$height-1) {
      $in->seek($offset + $y*$width*3 + $y*($width%4), 0);
      $in->read($buffer, $width*3);
      $raw_data .= $buffer;
    }
  } else {
    die qq{Unrecognised output image format "$image_format"!};
  }

  $in->close unless $params->{in_fh};

#  $params->{null_compress} = 1;
  my $num_levels = $params->{levels} || 5;

  my $output_data = '';
  my $output = IO::Scalar->new(\$output_data);

  {
    my $width = $width;
    my $height = $height;
    if ($params->{null_compress}) { $width *= 4; $height *= 4 }
  
    $output->print('RTS3',
                    pack(C3 => split //, $params->{format}),
                    pack(CLL => $num_levels, $width, $height));
  }

  my ($palette, $image_bits, $paletted_image);

  if ($params->{format} =~ /..3/) {
    if ($params->{format} =~ /00./) {
      $image_bits = 32;
    } elsif ($params->{format} =~ /.0./) {
      $image_bits = 16;
    } elsif ($params->{format} =~ /.1./) {
      $image_bits = 15;
    } elsif ($params->{format} =~ /.4./) {
      $image_bits = 12;
    } 

    my ($p32, $p16, $p15, $p15b, $p12) = (0,0,0,0,0);
    my $pal_offset = 40 + $num_levels*8;
    if ($image_bits == 32) { # includes 16 and 15
      $p32 =  $pal_offset;
      $p16 =  $pal_offset+256*4;
      $p15 =  $pal_offset+256*4+256*2;

      ($palette, $paletted_image) = image_to_ddt__generate_palette($raw_data);
      $palette = image_to_ddt__palette_32_to_32($palette)
               . image_to_ddt__palette_32_to_16($palette)
               . image_to_ddt__palette_32_to_15($palette);
    } else {
      $p16 =  $pal_offset  if $image_bits == 16;
      $p15 =  $p16 + 256*2 if $image_bits == 16;
      $p15b = $pal_offset  if $image_bits == 15;
      $p12 =  $pal_offset  if $image_bits == 12;

      ($palette, $paletted_image) = image_to_ddt__generate_palette($raw_data);
      $palette = image_to_ddt__palette_32_to_n($image_bits, $palette);
      $palette .= image_to_ddt__palette_32_to_15($palette) if $image_bits == 16;
    }

    $output->print(pack L6 => 256,$p32,$p16,$p15,$p15b,$p12);
  }


  my @images;

  for my $level (0..$num_levels-1) {

	my $image_data;
    if ($params->{format} =~ /..3/) {

      $image_data = image_to_ddt__resample_p($paletted_image, $level, $width, $height);

    } elsif ($params->{format} =~ /..7/) {

      $image_data = image_to_ddt__resample($raw_data, $level, $width, $height);
      $image_data =~ s/(.)../$1/sg; # convert RGB to 8-bit greyscale

    } else {
      $image_data = image_to_ddt__resample($raw_data, $level, $width, $height);
      my $alpha_data2 = image_to_ddt__resample_alpha($alpha_data, $level, $width, $height) if $got_alpha;

      my ($level_width, $level_height) = map $_<4?4:$_, map $_/(2**$level), $width, $height;
      $image_data = image_to_ddt__dxtc_compress($image_data, $alpha_data2, $params, $level_width, $level_height);
    }
    
    push @images, $image_data;
  }

  {
    my $offset = 16 + $num_levels*8;
    $offset += length($palette)+24 if defined $palette;
    for my $level (0..$num_levels-1) {
      $output->print(pack L2 => $offset, length($images[$level]));
      $offset += length($images[$level]);
    }
    $output->print($palette) if defined $palette;
    for my $level (0..$num_levels-1) {
      $output->print($images[$level]);
    }
  }

  my $out;
  if ($params->{out_fh}) {
    $out = $params->{out_fh};
    binmode $out unless ref $out eq 'IO::Scalar';
  } else {
    $out = IO::File->new("> $output_filename") or die "Error opening $output_filename for output: $!";
    binmode $out;
  }
  $out->print($output_data);
  $out->close unless $params->{out_fh};
}

sub image_to_ddt__dxtc_compress {

  my ($raw_data, $alpha_data, $params, $width, $height) = @_;

  die "ERROR: image dimensions ($width x $height) must be a multiple of 4!" if $width % 4 or $height % 4;

  my $image_data = '';

  if ($params->{null_compress}) { # null compression -- rubbish, but simple to implement

    if ($params->{format} =~ /..4/) {
      $raw_data =~ s/(...)/main::colour_to_16_rev($1)/seg;
    } elsif ($params->{format} =~ /..5/) {
      $raw_data =~ s/(...)/main::colour_to_15_rev($1)/seg;
    } elsif ($params->{format} =~ /..6/) {
      $raw_data =~ s/(...)/main::colour_to_12_rev($1)/seg;
    } else {
      die "Invalid pixel format ($params->{format})!";
    }

    if ($params->{format} =~ /..4/) {
      for my $y (0..$height-1) {
        for my $x (0..$width-1) {
          $image_data .= (substr($raw_data, 2*($y*$width+$x), 2) x 2) . "\0\0\0\0";
        }
      }
    } elsif ($params->{format} =~ /..5/) {
      my $alpha_bits;
      for my $y (0..$height-1) {
        for my $x (0..$width-1) {
          $alpha_bits = substr($alpha_data, $y*$width+$x, 1) eq "\0" ? "\0\0" : "\xFF\xFF";
          $image_data .= (substr($raw_data, 2*($y*$width+$x), 2) x 2) . $alpha_bits . "\0\0\0\0";
        }
      }
    } elsif ($params->{format} =~ /..6/) {
      my $alpha_bits;
      for my $y (0..$height-1) {
        for my $x (0..$width-1) {
          $alpha_bits = substr($alpha_data, $y*$width+$x, 1);
          vec($alpha_bits, 0, 4) = vec($alpha_bits, 1, 4);
          $image_data .= (substr($raw_data, 2*($y*$width+$x), 2) x 2)
                       . ($alpha_bits . $alpha_bits . "\0") x 4;
        }
      }
    } else {
      die "Invalid pixel format ($params->{format})!";
    }

  } else { # better compression, written largely in C for speed

    my ($orig_block, $colour1, $colour2, $pixels, $alpha_block, $alpha_bits);
    if ($params->{format} =~ /..4/) {
      for my $y (0..$height/4-1) {
        for my $x (0..$width/4-1) {
          $orig_block = substr($raw_data, 12*($y*$width+$x), 12)
                      . substr($raw_data, 3*((4*$y+1)*$width+4*$x), 12)
                      . substr($raw_data, 3*((4*$y+2)*$width+4*$x), 12)
                      . substr($raw_data, 3*((4*$y+3)*$width+4*$x), 12);
          ($colour1, $colour2, $pixels) = resrip::encode_block($orig_block, 0);
          $image_data .= pack SSL => $colour1, $colour2, $pixels;
        }
      }
    } elsif ($params->{format} =~ /..5/) {
      for my $y (0..$height/4-1) {
        for my $x (0..$width/4-1) {
          $orig_block = substr($raw_data, 12*($y*$width+$x), 12)
                      . substr($raw_data, 3*((4*$y+1)*$width+4*$x), 12)
                      . substr($raw_data, 3*((4*$y+2)*$width+4*$x), 12)
                      . substr($raw_data, 3*((4*$y+3)*$width+4*$x), 12);
          $alpha_block= substr($alpha_data, 4*($y*$width+$x), 4)
                      . substr($alpha_data, (4*$y+1)*$width+4*$x, 4)
                      . substr($alpha_data, (4*$y+2)*$width+4*$x, 4)
                      . substr($alpha_data, (4*$y+3)*$width+4*$x, 4);
          $alpha_bits = "\0\0";
          vec($alpha_bits, $_, 1) = vec($alpha_block, $_, 8) >> 7 for 0..15;
          ($colour1, $colour2, $pixels) = resrip::encode_block($orig_block, 1);
          $image_data .= pack(SS => $colour1, $colour2)
                      . $alpha_bits
                      . pack(L => $pixels);
        }
      }
    } elsif ($params->{format} =~ /..6/) {
      for my $y (0..$height/4-1) {
        for my $x (0..$width/4-1) {
          $orig_block = substr($raw_data, 12*($y*$width+$x), 12)
                      . substr($raw_data, 3*((4*$y+1)*$width+4*$x), 12)
                      . substr($raw_data, 3*((4*$y+2)*$width+4*$x), 12)
                      . substr($raw_data, 3*((4*$y+3)*$width+4*$x), 12);
          $alpha_block= substr($alpha_data, 4*($y*$width+$x), 4)
                      . substr($alpha_data, (4*$y+1)*$width+4*$x, 4)
                      . substr($alpha_data, (4*$y+2)*$width+4*$x, 4)
                      . substr($alpha_data, (4*$y+3)*$width+4*$x, 4);
          $alpha_bits = "\0\0\0\0\0\0\0\0";
          vec($alpha_bits, $_, 4) = vec($alpha_block, $_, 8) >> 4 for 0..15;
          ($colour1, $colour2, $pixels) = resrip::encode_block($orig_block, 2);
          $pixels = pack L => $pixels;
          $image_data .= pack(SS => $colour1, $colour2) .
                         substr($alpha_bits, 0, 2) . substr($pixels, 0, 1) .
                         substr($alpha_bits, 2, 2) . substr($pixels, 1, 1) .
                         substr($alpha_bits, 4, 2) . substr($pixels, 2, 1) .
                         substr($alpha_bits, 6, 2) . substr($pixels, 3, 1);
        }
      }
    } else {
      die "Invalid pixel format ($params->{format})!";
    }
  }
  return $image_data;
}

#sub image_to_ddt__generate_palette { # *VERY* BAD PALETTISING CODE!!
#  my ($bits, $raw_data) = @_;
#  if ($bits == 16) {
#    $raw_data =~ s/(...)/main::colour_to_16_rev($1)/seg;
#  } elsif ($bits == 15) {
#    $raw_data =~ s/(...)/main::colour_to_15_rev($1)/seg;
#  } elsif ($bits == 12) {
#    $raw_data =~ s/(...)/main::colour_to_12_rev($1)/seg;
#  } else {
#    die "Invalid number of bits ($bits)!";
#  }
#  my %colours;
#  ++$colours{$1} while $raw_data =~ /(..)/sg; # generate frequency table
#  my @top_colours = sort { $colours{$b} <=> $colours{$a} } keys %colours;
#  push @top_colours, "\0\0" while @top_colours < 256;
#  return join '', sort @top_colours[0..255];
#}


sub image_to_ddt__generate_palette {
  my ($raw_data) = @_;
  my $palette = "???" x 256;
  my $paletted = "?" x (length($raw_data)/3);
  resrip::create_palette(32, $raw_data, length($raw_data)/3, $palette, $paletted);
#{open O, '>pal.txt'; my @c=unpack 'C*', $palette; print O "JASC-PAL\n0100\n256\n";while (@c){print O join(' ', splice @c, 0, 3),"\n"; }close O;}
#print length $palette,"\n";
  $palette =~ s/(...)/$1\0/sg;
#print length $palette,"\n";
#{open O, '>pal2.txt'; my @c=unpack 'L*', $palette; print O join "\n", @c;close O;}
#{open O, '>pal3.txt'; my @c=unpack 'S*', image_to_ddt__palette_32_to_16($palette); print O join "\n", @c;close O;}

#  if ($bits == 16) {
 #   $palette =~ s/(...)/main::colour_to_16_rev($1)/seg;
 # } elsif ($bits == 15) {
 #   $palette =~ s/(...)/main::colour_to_15_rev($1)/seg;
 # } elsif ($bits == 12) {
 #   $palette =~ s/(...)/main::colour_to_12_rev($1)/seg;
 # } elsif ($bits == 32) {
 # 	$palette =~ s/(...)/$1\0/sg;
 # } else {
 #   die "Invalid number of bits ($bits)!";
 # }
  #die unless $bits == 32;
  return ($palette, $paletted);
}

#sub image_to_ddt__palette_16_to_32 {
#  return join '', map main::colour_16_to_bytes_rev($_)."\0", unpack 'S*' => shift;
#}
#sub image_to_ddt__palette_16_to_15 {
#  return join '', map main::colour_to_15_rev(main::colour_16_to_bytes_rev($_)), unpack 'S*' => shift;
#}
sub image_to_ddt__palette_32_to_32 {
  return join '', map main::colour_to_32(pack L => $_), unpack 'L*' => shift;
}
sub image_to_ddt__palette_32_to_16 {
  return join '', map main::colour_to_16(pack L => $_), unpack 'L*' => shift;
}
sub image_to_ddt__palette_32_to_15 {
  return join '', map main::colour_to_15(pack L => $_), unpack 'L*' => shift;
}
sub image_to_ddt__palette_32_to_12 {
  return join '', map main::colour_to_12(pack L => $_), unpack 'L*' => shift;
}
sub image_to_ddt__palette_32_to_n {
  return $_[0]==12 ? image_to_ddt__palette_32_to_12($_[1]) :
         $_[0]==15 ? image_to_ddt__palette_32_to_15($_[1]) :
         $_[0]==16 ? image_to_ddt__palette_32_to_16($_[1]) : die;
}

sub image_to_ddt__apply_palette { # not particularly good palette-applying code
  my ($raw_data, $palette, $bits, $alpha_bits) = @_;
  my (@palette_cache, $num_val, $closeness, $closest_val, $closest_col);
  my @palette = unpack 'S*' => substr($palette, 256*2);

  if ($bits == 32) {
    @palette = unpack 'S*' => substr($palette, 256*4, 256*2);
    $bits = 16;
  }

  if ($bits == 16) {
    $raw_data =~ s/(...)/main::colour_to_16_rev($1)/seg;
    $raw_data =~ s/(..)/
      $num_val = unpack S => $1;
      defined $palette_cache[$num_val] ? $palette_cache[$num_val] :
      do {
        $closest_val = 1e6;
        for (0..255) {
          $closeness = # compare as 5-bit values
             (($num_val >> 11) - ($palette[$_] >> 11)) ** 2
           + ((($num_val >> 6) & 31) - (($palette[$_] >> 6) & 31)) ** 2
           + (($num_val & 31) - ($palette[$_] & 31)) ** 2;
          if ($closeness < $closest_val) {
            $closest_col = $_; $closest_val = $closeness;
            last if $closeness == 0;
          }
        }
        $palette_cache[$num_val] = pack C => $closest_col;
        $palette_cache[$num_val]; # (return value)
      }
    /seg;

  } elsif ($bits == 15) {
    $raw_data =~ s/(...)/main::colour_to_15_rev($1)/seg;
    $raw_data =~ s/(..)/
      $num_val = unpack S => $1;
      defined $palette_cache[$num_val] ? $palette_cache[$num_val] :
      do {
        $closest_val = 1e6;
        for (0..255) {
          $closeness = # compare as 5-bit values
             (($num_val >> 10) - ($palette[$_] >> 10)) ** 2
           + ((($num_val >> 5) & 31) - (($palette[$_] >> 5) & 31)) ** 2
           + (($num_val & 31) - ($palette[$_] & 31)) ** 2;
          if ($closeness < $closest_val) {
            $closest_col = $_; $closest_val = $closeness;
            last if $closeness == 0;
          }
        }
        $palette_cache[$num_val] = pack C => $closest_col;
        $palette_cache[$num_val]; # (return value)
      }
    /seg;

  } elsif ($bits == 12) { # includes alpha
    $raw_data =~ s/(...)/main::colour_to_12_rev($1)/seg;
    $raw_data =~ s/(..)/
      $num_val = unpack S => $1;
      defined $palette_cache[$num_val] ? $palette_cache[$num_val] :
      do {
        $closest_val = 1e6;
        for (0..255) {
          $closeness = # compare as 4-bit values
             (($num_val >> 8) - ($palette[$_] >> 8)) ** 2
           + ((($num_val >> 4) & 15) - (($palette[$_] >> 4) & 15)) ** 2
           + (($num_val & 15) - ($palette[$_] & 15)) ** 2;
          if ($closeness < $closest_val) {
            $closest_col = $_; $closest_val = $closeness;
            last if $closeness == 0;
          }
        }
        $palette_cache[$num_val] = pack C => $closest_col;
        $palette_cache[$num_val]; # (return value)
      }
    /seg;

    my $p = 0;
    my $alpha;
    $raw_data =~ s/(....)/
      $alpha = substr($alpha_bits, 4*$p++, 4);
      pack(C2 => (vec($alpha, 1, 8) >> 4) | (vec($alpha, 0, 8) & 240),
                 (vec($alpha, 3, 8) >> 4) | (vec($alpha, 2, 8) & 240)) . $1;
    /seg;
   
  } else {
    die "Invalid number of bits ($bits)!";
  }
  return $raw_data;
}

sub image_to_ddt__resample { # very bad resampling code!
  my ($raw_data, $level, $width, $height) = @_;
  my $return;
  if ($level == 0) {
    return $raw_data;
  } else {
    for my $y (0..$height-1) {
      next if $y % (2**$level);
      for my $x (0..$width-1) {
        next if $x % (2**$level);
        $return .= substr($raw_data, 3*($y*$width+$x), 3);
      }
    }
    $return = substr($raw_data, 0, 3) if not length $return; # never be <1x1 pixels
    return $return;
  }
}
sub image_to_ddt__resample_alpha {
  my ($raw_data, $level, $width, $height) = @_;
  my $return;
  if ($level == 0) {
    return $raw_data;
  } else {
    for my $y (0..$height-1) {
      next if $y % (2**$level);
      for my $x (0..$width-1) {
        next if $x % (2**$level);
        $return .= substr($raw_data, $y*$width+$x, 1);
      }
    }
    $return = substr($raw_data, 0, 1) if not length $return; # never be <1x1 pixels
    return $return;
  }
}
sub image_to_ddt__resample_p { image_to_ddt__resample_alpha(@_) }


################################################

sub brg_to_3ds {
  my ($input_filename, $output_filename, $params) = @_;
  my $in;
  if ($params->{in_fh}) {
    $in = $params->{in_fh};
    binmode $in unless ref $in eq 'IO::Scalar';
  } else {
    $in = IO::File->new("< $input_filename") or die "Error opening $input_filename for input: $!";
    binmode $in;
  }

  my $data = brg_to_3ds__read_brg($in);

  $in->close unless $params->{in_fh};

  # return (
  #   $data->{num_meshes},
  #   $data->{num_frames},
  #   $data->{mesi_0_num_faces},
  #   $data->{mesi_0_num_vertices},
  # );

  unless ($params and $params->{no_output}) {
    my $out;
    if ($params->{out_fh}) {
      $out = $params->{out_fh};
      binmode $out unless ref $out eq 'IO::Scalar';
    } else {
      $out = IO::File->new("> $output_filename") or die "Error opening $output_filename for output: $!";
      binmode $out;
    }

    my $output = brg_to_3ds__output_3ds($data, {%$params, brg_filename=>$output_filename});
    $out->print($output->{data});

    if ($params->{texturesbar}) {
      foreach my $name (keys %{$output->{ddts}}) {
        my $output_file = File::Spec->catpath((File::Spec->splitpath($output_filename))[0,1], $name);
        next if -e $output_file;

        # ddt_to_image($params->{do_ddt}."/$output->{ddts}{$name}.ddt",
        #              $output_file);

        my $bar_data = bar_extract_file($params->{texturesbar}, "$output->{ddts}{$name}.ddt");
        my $barfile_fh = IO::Scalar->new(\$bar_data);
        ddt_to_image(undef, $output_file, { in_fh => $barfile_fh, imageformat => 'tga' });
      }
    }

#    $out->print(brg_to_3ds__dump_text($data));

    $out->close unless $params->{out_fh};

    return { ddts => $output->{ddts} };
  }
}

sub brg_to_3ds__squeeze_name { # squash to 16 bytes (for material names)
  my ($name) = @_;

  my @removals = ('a'..'z', qw(
   building map special world head standard cavalry attachments attachment
   infantry hero iron animal scenario sfx villager copper archer siege
   shield bronze naval statue horse trade caravan giant ui all sword male
   flag of banners female relic axe ram wonder hypaspist jarl settlement
   wonder01 huskarl axeman accessories selected construction age2 toxotes
   shared01 hippikon skyboxsunset hoplite raiding water skyboxstormy spear
   frost skyboxwinter wonder02 wall skyboxbluesky prodromos stone isis cargo
   age4 set ballista myrmidon man ulfsark poseidon spearman cape garrison
   chimera hades elephant hetairoi theseus jason burnt cataphract fish kamos
   ra heracles greek war troy age3 throwing polyphemus fortress slinger tent
   passage einheriar color), 0..9);

  $name =~ s/_/ /g; # split underscored names
  $name =~ s/([a-z])([A-Z])/$1 $2/g; # split capitalisationally split names

  my @words = split /\s+/, $name;
  foreach my $rem (@removals) { # trim common words until it's short enough
    last if length(join '', @words) <= 16;
    @words = grep lc($_) ne $rem, @words;
  }
  $name = join '', map ucfirst(lc $_), @words; # convert qw(abc def) to 'AbcDef'

  # if it's still too long...
  1 while length $name > 16 and $name =~ s/[0-9]//; # remove numbers
  1 while length $name > 16 and $name =~ s/[aeiou]//; # remove vowels
  $name = substr($name, 0, 16) if length $name > 16; # if all else fails
  return $name;
}

sub brg_to_3ds__output_3ds {
  my %data = %{shift()};
  my $params = shift;

# ought to rescale by ~100x

  my $main_3ds = new File::3DS;
  my $edit_chunk = $main_3ds->new(type => 'EDIT3DS', data => { version => 2 } );

  my %materials;

  my %used_names;
  my %ddts; # store a list of necessary DDT->TGA conversions

  for my $mat_num (0..$data{num_materials}-1) {

    my $mat_id = $data{"mtrl_${mat_num}_id"};
    my $mat_name = $data{"mtrl_${mat_num}_name"};

    my $name = brg_to_3ds__squeeze_name(length $mat_name ? $mat_name : 'unnamed');
    if ($used_names{$name}++) {
      my $n = -1; my $new_name;
      do { $new_name = substr($name, 0, 16-length(++$n)) . $n } while $used_names{$new_name}++;
      $name = $new_name;
    }

    $materials{$mat_id} = { # BRG's arbitrary ID
      name => $name,
      full_name => $mat_name,
      faces => [],
      vertices => []
    };

    # just use "mat_nnnn.bmp" files (but get confused with the 'low' models)
    my $filename = 'mat_'.sprintf('%04d', $mat_id).'.tga';

    $ddts{$filename} = $mat_name if length $mat_name;
    # some materials have no name (even in materials.lib) -- why??
  

    my $material_chunk = $edit_chunk->new(type => 'EDIT_MATERIAL', data => { } );
    $material_chunk->new(type => 'MAT_NAME01', data => { name => $name } );
    my $colour_chunk = $material_chunk->new(type => 'MAT_DIFFUSE', data => { } );
    $colour_chunk->new(type => 'COL_RGBBYTE', data => { r => 0, g => 0, b => 0 } );

    if (length $mat_name) {
      my $texture_chunk = $material_chunk->new(type => 'MAT_TEXTURE01', data => { } );
      $texture_chunk->new(type => 'MAP_FILENAME', data => { filename => $filename } );

# WARNING: Lots of bad guessing here!
      unless (defined $params->{brg_filename}
             and $params->{brg_filename} =~ /building/i) { # buildings generally don't have alpha
        # if (matlib_alpha($params->{matlib}, $mat_id)) { # <-- doesn't work
        if ($mat_num == 0) { # seems to be the first that has alpha, and the second with player-colours
          my $opacity_chunk = $material_chunk->new(type => 'MAT_OPACITY', data => { } );
          $opacity_chunk->new(type => 'MAP_FILENAME', data => { filename => $filename } );
        }
      }
    }
  }

  { # Store with materials the faces/vertices they apply to
    my $face_id = 0;
    push @{$materials{$_}{faces}}, $face_id++ foreach unpack 's*' => $data{mesi_0_face_mats};

    my $vertex_id = 0;
    push @{$materials{$_}{vertices}}, $vertex_id++ foreach unpack 's*' => $data{mesi_0_vertex_mats};
  }

  my @texture_coords = ([0, 0]) x $data{mesi_0_num_vertices}; # default texture coords

  my @faces;
  push @faces, [ unpack(s3 => substr($data{mesi_0_faces}, $_*6, 6)), 0b1111 ]
    for 0..$data{mesi_0_num_faces}-1;

  my @smoothing;
  push @smoothing, pack L => 1
    for 0..$data{mesi_0_num_faces}-1; # stick everything in smoothing group 1

  for my $frame_id (0..$data{num_meshes}-1) {

    my $object_chunk = $edit_chunk->new(type => 'EDIT_OBJECT', data => { name => 'Frame'.sprintf('%03d',$frame_id) } );
    my $mesh_chunk = $object_chunk->new(type => 'OBJ_TRIMESH', data => { } );

    if (length $data{"mesi_${frame_id}_texture_coords"}) { # update texture coordinates if they're provided
      @texture_coords = ();
      push @texture_coords, [ unpack_float16(substr($data{"mesi_${frame_id}_texture_coords"}, $_*4, 4)) ]
        for 0..$data{"mesi_${frame_id}_num_vertices"}-1;
    }

    my @vertices;
    do { my @vertex = unpack_float16(substr($data{"mesi_${frame_id}_vertices"}, $_*6, 6));
         push @vertices, [ $vertex[0], -$vertex[2], $vertex[1] ] }
      for 0..$data{"mesi_${frame_id}_num_vertices"}-1;

#    my @normals; # ??? weird values :-(
#    do { my @normal = unpack_float16(substr($data{"mesi_${frame_id}_normals"}, $_*6, 6));
#         push @normals, [ $normal[0], -$normal[2], $normal[1] ] }
#      for 0..$data{"mesi_${frame_id}_num_vertices"}-1;
#
#print join "\n", map join(", ", @$_), @vertices;
#print "\n\n";
#print join "\n", map join(", ", @$_), @normals;

    $mesh_chunk->new(type => 'TRI_VERTEXL', data => { vertices => [@vertices] } );
    $mesh_chunk->new(type => 'TRI_MAPPINGCOORDS', data => { coords => [@texture_coords] } );

    my $faces_chunk = $mesh_chunk->new(type => 'TRI_FACEL1', data => { faces => [@faces] } );

    $faces_chunk->new(type => 'TRI_SMOOTH', data => { faces => [@smoothing] } );

    foreach my $mat_num (sort keys %materials) {
      $faces_chunk->new(type => 'TRI_MATERIAL', data => { name => $materials{$mat_num}{name}, faces => [@{$materials{$mat_num}{faces}}] } );
    }
  }

  return { data => $main_3ds->write(scale => 1), ddts => \%ddts };

}

sub brg_to_3ds__dump_text {
  my %data = %{shift()};
  my $return = '';
  my $out = IO::Scalar->new(\$return);

  print $out "Materials: $data{num_materials}\n";
  print $out "Meshes: $data{num_meshes}\n";
  print $out "Frames: $data{num_frames}\n" if $data{num_meshes}-1;


  for my $m (0..$data{num_materials}-1) {
    print $out "MTRL $m:\n";
    print $out "  id:   ", $data{"mtrl_${m}_id"}, "\n";
    print $out "  name: ", $data{"mtrl_${m}_name"}, "\n";
    print $out "  unknowns:
    01: ",join(' ', unpack 'C*', $data{"mtrl_${m}_unknown01"}), "
    02: ",join(' ', unpack 'C*', $data{"mtrl_${m}_unknown02"}), "
    03: ",join(' ', unpack 'C*', $data{"mtrl_${m}_unknown03"}), "
    04: ",join(' ', unpack 'C*', $data{"mtrl_${m}_unknown04"}), "
";
    print $out "    05: ",join(' ', unpack 'C*', $data{"mtrl_${m}_unknown05"}), "
"     if exists $data{"mtrl_${m}_unknown05"};

    if (exists $data{"mtrl_${m}_num_sfx"}) {
      print $out "SFX: ", $data{"mtrl_${m}_num_sfx"}, "\n";
      for my $m (0..$data{"mtrl_${m}_num_sfx"}-1) {
        print $out "SFX $m:\n";
        print $out "  id:   ", $data{"mtrl_${m}_sfx_${m}_id"}, "\n";
        print $out "  name: ", $data{"mtrl_${m}_sfx_${m}_name"}, "\n";
      }
    }
  }

  for my $m (0..$data{num_meshes}-1) {
    print $out "MESI $m:\n";
    print $out "  Faces: ",$data{"mesi_${m}_num_faces"},"\n";
    print $out "  Vertices: ",$data{"mesi_${m}_num_vertices"},"\n";
    if (1) {
      print $out "  unknowns:
    [01: ",join(' ', unpack 'C*', $data{unknown01}), "
    [02: ",join(' ', unpack 'C*', $data{unknown02}), "
    [03: ",join(' ', unpack 'C*', $data{unknown03});
      print $out "
    [04: ",join(' ', unpack 'f*', $data{unknown04})
        if exists $data{unknown04};
      print $out "
    01b: ",$data{"mesi_${m}_unknown01b"},"
    02:  ",$data{"mesi_${m}_unknown02"},"
    03:  ",join(' ', unpack 'H*', $data{"mesi_${m}_unknown03"}),"
    04:  ",join(' ', unpack 'C*', $data{"mesi_${m}_unknown04"}),"
    05:  ",$data{"mesi_${m}_unknown05"},"
    06:  ",$data{"mesi_${m}_unknown06"},"
    07:  ",join(' ', unpack 'f*', $data{"mesi_${m}_unknown07"});

    if (1) {
#      print $out "
#    vert:",join(' ', unpack 'H*', $data{"mesi_${m}_vertices"}),"
#    norm:",join(' ', unpack 'H*', $data{"mesi_${m}_normals"});
#      print $out "
#    text:",join(' ', unpack 'H*', $data{"mesi_${m}_texture_coords"})
#      if exists $data{"mesi_${m}_texture_coords"};
      print $out "
    vert:",join(' ', unpack_float16($data{"mesi_${m}_vertices"})),"
    norm:",join(' ', unpack_float16($data{"mesi_${m}_normals"}));
#"
#    norm:",join(' ', unpack 'f*', $data{"mesi_${m}_normals"});
      print $out "
    text:",join(' ', unpack_float16($data{"mesi_${m}_texture_coords"}))
      if exists $data{"mesi_${m}_texture_coords"};

      print $out "
    fcmt:",join(' ', unpack 'H*', $data{"mesi_${m}_face_mats"}),"
    face:",join(' ', unpack 'H*', $data{"mesi_${m}_faces"}),"
    vxmt:",join(' ', unpack 'H*', $data{"mesi_${m}_vertex_mats"})
      if $m==0;
    }

      print $out "
    08:  ",join('', map "\n      ".unpack('H*', $_), @{$data{"mesi_${m}_unknown08"}})
      if exists $data{"mesi_${m}_unknown08"};

      print $out "
    09:  ",join('', map "\n      ".unpack('H*', $_), @{$data{"mesi_${m}_unknown09"}}), "
    09a: ", $data{"mesi_${m}_unknown09a"}, "
    09e: ", $data{"mesi_${m}_unknown09e"};

      print $out "
    09b: ", unpack('H*', $data{"mesi_${m}_unknown09b"}), "
    09c: ", unpack('l', $data{"mesi_${m}_unknown09c"}), "
    09d: ", unpack('l', $data{"mesi_${m}_unknown09d"})
      if exists $data{"mesi_${m}_unknown09b"};

      print $out "
    0a:  ",join('', map "\n      ".unpack('H*', $_), @{$data{"mesi_${m}_unknown0a"}})
      if exists $data{"mesi_${m}_unknown0a"};

      print $out "
    c1:  ",$data{"mesi_${m}_count1"},"
    c2:  ",$data{"mesi_${m}_count2"},"
    10:  ",join(' ', unpack 'C*', $data{"mesi_${m}_unknown10"}),"
    11:  ",join('', map "\n      ".unpack('H*', $_), @{$data{"mesi_${m}_unknown11"}}),"
    12:  ",join('', map "\n      ".unpack('H*', $_), @{$data{"mesi_${m}_unknown12"}}),"
    13:  ", join(' ', unpack 'C*', $data{"mesi_${m}_unknown13"}),"
    14:  ", join(' ', unpack 'f*', $data{"mesi_${m}_unknown14"})
        if exists $data{"mesi_${m}_count1"};

      print $out "\n";
    }
  }

  return $return;
}

sub brg_to_3ds__read_brg {
  my $in = shift;
  my %return;
  my $buffer;

  my $die = sub { die @_ };
#  my $die = sub { warn @_ };
#  $SIG{__DIE__} = sub { print brg_to_3ds__dump_text(\%return) };
#  $SIG{__WARN__} = sub { die @_ };
#  $SIG{__WARN__} = sub { $|=1 };

  $in->read($buffer, 4+4+4+4+4+8);
  @return{qw(file_head unknown01 num_materials unknown02 num_meshes unknown03)}
   = unpack( 'a4       a4        l             a4        l          a8',    $buffer);

  $return{file_head} eq 'BANG' or $die->("Invalid file header ($return{file_head})!"), return \%return;

# aom.exe says: chunks are MESI, AFRM(?!) or ASET

  if ($return{num_meshes} > 1 or ( $return{unknown03} eq "\0\0\0\0\3\0\0\0" and do {
    $in->read($buffer, 4); $in->seek(-4, 1); $buffer eq 'ASET';
  } )) { # not the nicest method...
    $in->read($buffer, 4+4+24);
    @return{qw(aset_head num_frames unknown04)}
     = unpack( 'a4       l          a24',  $buffer);

    $return{aset_head} eq 'ASET' or $die->("Invalid ASET header ($return{aset_head})!"), return \%return;

    $return{num_frames} == $return{num_meshes} or $die->("num_frames != num_meshes ($return{num_frames} != $return{num_meshes})!"), return \%return;
  }

  for my $mesh_id (0..$return{num_meshes}-1) {

    my $s = sub{wantarray?map"mesi_${mesh_id}_$_",@_:"mesi_${mesh_id}_$_[0]"};

    $in->read($buffer, 4+4+2+2+4+36+6+1+1+24);

    @return{$s->(qw(head meshformat unknown01b num_vertices num_faces unknown02 unknown03 unknown04 unknown05 unknown06 unknown07))}
     = unpack (    'a4   s          s          s            s         l         a36       a6        c         C         a24',    $buffer);

    $return{$s->('head')} eq 'MESI' or $die->("Invalid MESI header (@{[$return{$s->('head')}]})"), return \%return;

    ### return \%return; # for quickly getting some stats about lots of files

    $in->read($buffer, 6 * $return{$s->('num_vertices')});
    $return{$s->('vertices')} = $buffer;
#    $return{$s->('vertices')} = [ map [ unpack_float16(substr($buffer, $_*6, 6)) ], 0..$return{$s->('num_vertices')}-1 ];

    $in->read($buffer, 6 * $return{$s->('num_vertices')});
    $return{$s->('normals')} = $buffer;
#    $return{$s->('normals')} = [ map [ unpack_float16(substr($buffer, $_*6, 6)) ], 0..$return{$s->('num_vertices')}-1 ];

      if (
          ( $return{$s->('unknown05')} == 64 )#and $return{$s->('unknown06')} == 1)
        or
         ($return{num_meshes} > 1 and
           ( $return{$s->('unknown05')} == 64
          or ($return{$s->('unknown05')} == 97)
          or ($return{$s->('unknown05')} == 66 and $return{$s->('unknown06')} == 4)
          or ($return{$s->('unknown05')} == 66 and $return{$s->('unknown06')} == 5)
           )
          and ($return{$s->('unknown02')} != 256)
         )) {
        $return{$s->('texture_coords')} = '';
      } else {
        $in->read($buffer, 4 * $return{$s->('num_vertices')});
        $return{$s->('texture_coords')} = $buffer;
#        $return{$s->('texture_coords')} = [ map [ unpack_float16(substr($buffer, $_*4, 4)) ], 0..$return{$s->('num_vertices')}-1 ];
      }


    if ($mesh_id == 0) {

      $in->read($buffer, 2 * $return{$s->('num_faces')});
      $return{$s->('face_mats')} = $buffer;
#      $return{$s->('face_mats')} = [ unpack 's*' => $buffer ];

      $in->read($buffer, 6 * $return{$s->('num_faces')});
      $return{$s->('faces')} = $buffer;
#      $return{$s->('faces')} = [ map [ unpack s3 => substr($buffer, $_*6, 6) ], 0..$return{$s->('num_faces')}-1 ];

      $in->read($buffer, 2 * $return{$s->('num_vertices')});
      $return{$s->('vertex_mats')} = $buffer;
#      $return{$s->('vertex_mats')} = [ unpack 's*' => $buffer ];
    }

    $in->read($buffer, 24);
    $return{$s->('unknown09')} = [ map substr($buffer, $_*4, 4), 0..5 ];

    $in->read($buffer, 2);
    $return{$s->('unknown09a')} = unpack s => $buffer;
    $in->read($buffer, 2);
    $return{$s->('unknown09e')} = unpack s => $buffer;

    if (not $return{$s->('unknown09a')}) {
      $in->read($buffer, 12);
      @return{$s->(qw(unknown09b unknown09c unknown09d))} =
        map substr($buffer, $_*4, 4), 0..2;
    }


    if ($return{$s->('unknown06')} == 200
     or $return{$s->('unknown06')} == 204
     or $return{$s->('unknown06')} == 72
     or ($return{$s->('unknown06')} == 76 and $return{$s->('unknown05')} != 98)
     or $return{$s->('unknown05')} == 97
    ) { # ?!?!
      $in->read($buffer, 4 * $return{$s->('num_vertices')});
      $return{$s->('unknown0a')} = [ map substr($buffer, $_*4, 4), 0..$return{$s->('num_vertices')}-1 ];
    }

    if ($return{$s->('unknown06')} & 1) {
      $in->read($buffer, 6);
      @return{$s->(qw(count1 count2 unknown10))}
       = unpack (    's      s      s',    $buffer);

      $in->read($buffer, 36 * $return{$s->('count1')});
      $return{$s->('unknown11')} = [ map substr($buffer, $_*36, 36), 0..$return{$s->('count1')}-1 ];

      $in->read($buffer, 8 * $return{$s->('count2')});
      $return{$s->('unknown12')} = [ map substr($buffer, $_*8, 8), 0..$return{$s->('count2')}-1 ];

      my $count = 0;
      $count += unpack(L => $_) foreach @{$return{$s->('unknown12')}};

      $in->read($buffer, $count);
      $return{$s->('unknown13')} = $buffer;

      if (exists $return{$s->('unknown09c')} and unpack l => $return{$s->('unknown09c')}) {
        $in->read($buffer, unpack(l => $return{$s->('unknown09c')})*4);
        $return{$s->('unknown14')} = $buffer;
      } else {
        $return{$s->('unknown14')} = '';
      }
    }
  }

  for my $matr_id (0..$return{num_materials}-1) {
    my $s = sub{wantarray?map"mtrl_${matr_id}_$_",@_:"mtrl_${matr_id}_$_[0]"};

    $in->read($buffer, 4+2+10+4+36+12);
    @return{$s->(qw(head id unknown01 name_length unknown02 unknown03))}
     = unpack (    'a4   s  a10       l           a36       a12',    $buffer);

# maybe... if unknown01[4] & 4, the 'alpha' channel means player colour (and matlib's unknown03 is 4)

    $return{$s->('head')} eq 'MTRL' or $die->("Invalid MTRL header (@{[$return{$s->('head')}]})"), return \%return;

    $in->read($buffer, $return{$s->('name_length')});
    $return{$s->('name')} = $buffer;

    $in->read($buffer, 4);
    $return{$s->('unknown04')} = $buffer;

    if ((unpack SSS => $return{$s->('unknown01')})[2] & 2) {
      $in->read($buffer, 4);
      $return{$s->('unknown05')} = $buffer;
    }

    if ( # erk... nasty code! [since I don't know how to detect whether there's an SFX]
        $return{$s->('id')} == 591
     or $return{$s->('id')} == 602
     or $return{$s->('id')} == 803
     or $return{$s->('id')} == 804
     or $return{$s->('id')} == 876
     or $return{$s->('id')} == 877
     or ($return{$s->('name')} eq 'Naval N Kraken' and (
           $return{$s->('id')} == 5 or
           $return{$s->('id')} == 17 or
           $return{$s->('id')} == 18 or
           $return{$s->('id')} == 19 or
           $return{$s->('id')} == 20 or
           $return{$s->('id')} == 21 or
           $return{$s->('id')} == 23 or
           $return{$s->('id')} == 25 or
           $return{$s->('id')} == 31))
     or ($return{$s->('name')} eq 'Special N Battle Boar' and (
           $return{$s->('id')} == 3 or
           $return{$s->('id')} == 4))
       ) {

      $in->read($buffer, 1);
      $return{$s->('num_sfx')} = unpack c => $buffer;
  
      for my $sfx_id (0..$return{$s->('num_sfx')}-1) {
        my $s = sub{wantarray?map"mtrl_${matr_id}_sfx_${sfx_id}_$_",@_:"mtrl_${matr_id}_sfx_${sfx_id}_$_[0]"};
  
        $in->read($buffer, 1);
        $return{$s->('id')} = unpack c => $buffer;
  
        $in->read($buffer, 2);
        $return{$s->('name_length')} = unpack s => $buffer;
  
        $in->read($buffer, $return{$s->('name_length')});
        $return{$s->('name')} = $buffer;
  
        die "Invalid SFX name @{[$return{$s->('name')}]}!" unless $return{$s->('name')} =~ /SFX/;
      }
    }
  }

  if (not $in->eof) {
    if (exists $return{mtrl_2_name} and $return{mtrl_2_name} eq 'Special N Raven') {
      # "special n raven_birth.brg" has 4099 bytes of seemingly useless repeated data
    } elsif (exists $return{mtrl_1_name} and $return{mtrl_1_name} eq 'world a breastplate map') {
      # "world g weapons 4.brg" has 98 bytes of incorrect duplicate MTRL data
    } else {
#      print "\n\n\n", $in->read($buffer, 1e6), ":\n", join('', unpack 'H*', $buffer);
      die "Not EOF!";
    }
  }

  return \%return;
}




sub threeds_to_brg {
  my ($input_filename, $output_filename, $params) = @_;
  my $in;
  if ($params->{in_fh}) {
    $in = $params->{in_fh};
  } else {
    $in = IO::File->new("< $input_filename") or die "Error opening $input_filename for input: $!";
  }
  my $update_data = join '', $in->getlines;
  $in->close;
  $update_data =~ s/.*\*\*START\*\*\n(.*)\n\*\*END\*\*/$1/s or die "Cannot find **START** and **END** lines in update file data";

  my $data;
  if ($params->{data}) {
    $data = $params->{data};
  } else {
    my $brg = IO::File->new("< $params->{brg}") or die "Error opening BRG $params->{brg} for input: $!";
    binmode $brg;
    $data = brg_to_3ds__read_brg($brg);
    $brg->close;
  }

  my $new_data = threeds_to_brg__patch_data($update_data, $data);

  my $out;
  if ($params->{out_fh}) {
    $out = $params->{out_fh};
    binmode $out unless ref $out eq 'IO::Scalar';
  } else {
    $out = IO::File->new("> $output_filename") or die "Error opening $output_filename for output: $!";
    binmode $out;
  }
  threeds_to_brg__write_brg($out, $new_data);

  $out->close;
}

sub threeds_to_brg__write_brg { # should theoretically produce an exact copy if run from the output of brg_to_3ds__read_brg
  my $out = shift;
  my %data = %{$_[0]};

  my $die = sub { die @_ };

#  $in->read($buffer, 4+4+4+4+4+8);
#  @return{qw(file_head unknown01 num_materials unknown02 num_meshes unknown03)}
#   = unpack( 'a4       a4        l             a4        l          a8',    $buffer);

  $out->print(pack( 'a4       a4        l             a4        l          a8',
           @data{qw(file_head unknown01 num_materials unknown02 num_meshes unknown03)}));


#  $return{file_head} eq 'BANG' or $die->("Invalid file header ($return{file_head})!"), return \%return;


#  if ($return{num_meshes} > 1 or ( $return{unknown03} eq "\0\0\0\0\3\0\0\0" and do {
#    $in->read($buffer, 4); $in->seek(-4, 1); $buffer eq 'ASET';
#  } )) { # not the nicest method...
#    $in->read($buffer, 4+4+24);
#    @return{qw(aset_head num_frames unknown04)}
#     = unpack( 'a4       l          a24',  $buffer);
#
#    $return{aset_head} eq 'ASET' or $die->("Invalid ASET header ($return{aset_head})!"), return \%return;
#
#    $return{num_frames} == $return{num_meshes} or $die->("num_frames != num_meshes ($return{num_frames} != $return{num_meshes})!"), return \%return;
#  }

  if (defined $data{aset_head} and $data{aset_head} eq 'ASET') {
    $out->print(pack( 'a4       l          a24',
             @data{qw(aset_head num_frames unknown04)}));
  }


  for my $mesh_id (0..$data{num_meshes}-1) {

    my $s = sub{wantarray?map"mesi_${mesh_id}_$_",@_:"mesi_${mesh_id}_$_[0]"};

#    $in->read($buffer, 4+4+2+2+4+36+6+1+1+24);
#
#    @return{$s->(qw(head meshformat unknown01b num_vertices num_faces unknown02 unknown03 unknown04 unknown05 unknown06 unknown07))}
#     = unpack (    'a4   s          s          s            s         l         a36       a6        c         C         a24',    $buffer);

    $out->print(pack (    'a4   s          s          s            s         l         a36       a6        c         C         a24',
             @data{$s->(qw(head meshformat unknown01b num_vertices num_faces unknown02 unknown03 unknown04 unknown05 unknown06 unknown07))}));

#    $return{$s->('head')} eq 'MESI' or $die->("Invalid MESI header (@{[$return{$s->('head')}]})"), return \%return;


#    $in->read($buffer, 6 * $return{$s->('num_vertices')});
#    $return{$s->('vertices')} = $buffer;
##    $return{$s->('vertices')} = [ map [ unpack_float16(substr($buffer, $_*6, 6)) ], 0..$return{$s->('num_vertices')}-1 ];

    $out->print($data{$s->('vertices')});

#    $in->read($buffer, 6 * $return{$s->('num_vertices')});
#    $return{$s->('normals')} = $buffer;
##    $return{$s->('normals')} = [ map [ unpack_float16(substr($buffer, $_*6, 6)) ], 0..$return{$s->('num_vertices')}-1 ];

    $out->print($data{$s->('normals')});

#      if (
#          ( $return{$s->('unknown05')} == 64 )#and $return{$s->('unknown06')} == 1)
#        or
#         ($return{num_meshes} > 1 and
#           ( $return{$s->('unknown05')} == 64
#          or ($return{$s->('unknown05')} == 97)
#          or ($return{$s->('unknown05')} == 66 and $return{$s->('unknown06')} == 4)
#          or ($return{$s->('unknown05')} == 66 and $return{$s->('unknown06')} == 5)
#           )
#          and ($return{$s->('unknown02')} != 256)
#         )) {
#        $return{$s->('texture_coords')} = '';
#      } else {
#        $in->read($buffer, 4 * $return{$s->('num_vertices')});
#        $return{$s->('texture_coords')} = $buffer;
##        $return{$s->('texture_coords')} = [ map [ unpack_float16(substr($buffer, $_*4, 4)) ], 0..$return{$s->('num_vertices')}-1 ];
#      }

    $out->print($data{$s->('texture_coords')});


    if ($mesh_id == 0) {

#      $in->read($buffer, 2 * $return{$s->('num_faces')});
#      $return{$s->('face_mats')} = $buffer;
##      $return{$s->('face_mats')} = [ unpack 's*' => $buffer ];
      $out->print($data{$s->('face_mats')});

#      $in->read($buffer, 6 * $return{$s->('num_faces')});
#      $return{$s->('faces')} = $buffer;
##      $return{$s->('faces')} = [ map [ unpack s3 => substr($buffer, $_*6, 6) ], 0..$return{$s->('num_faces')}-1 ];
      $out->print($data{$s->('faces')});

#      $in->read($buffer, 2 * $return{$s->('num_vertices')});
#      $return{$s->('vertex_mats')} = $buffer;
##      $return{$s->('vertex_mats')} = [ unpack 's*' => $buffer ];
      $out->print($data{$s->('vertex_mats')});
    }

#    $in->read($buffer, 24);
#    $return{$s->('unknown09')} = [ map substr($buffer, $_*4, 4), 0..5 ];
    $out->print(join '', @{$data{$s->('unknown09')}});

#    $in->read($buffer, 2);
#    $return{$s->('unknown09a')} = unpack s => $buffer;
    $out->print(pack s => $data{$s->('unknown09a')});
#    $in->read($buffer, 2);
#    $return{$s->('unknown09e')} = unpack s => $buffer;
    $out->print(pack s => $data{$s->('unknown09e')});

    if (not $data{$s->('unknown09a')}) {
#      $in->read($buffer, 12);
#      @return{$s->(qw(unknown09b unknown09c unknown09d))} =
#        map substr($buffer, $_*4, 4), 0..2;
      $out->print(@data{$s->(qw(unknown09b unknown09c unknown09d))});
    }


#    if ($return{$s->('unknown06')} == 200
#     or $return{$s->('unknown06')} == 204
#     or $return{$s->('unknown06')} == 72
#     or ($return{$s->('unknown06')} == 76 and $return{$s->('unknown05')} != 98)
#     or $return{$s->('unknown05')} == 97
#    ) { # ?!?!
#      $in->read($buffer, 4 * $return{$s->('num_vertices')});
#      $return{$s->('unknown0a')} = [ map substr($buffer, $_*4, 4), 0..$return{$s->('num_vertices')}-1 ];
#    }
    if ($data{$s->('unknown0a')}) {
      $out->print(@{$data{$s->('unknown0a')}});
    }

#    if ($return{$s->('unknown06')} & 1) {
#      $in->read($buffer, 6);
#      @return{$s->(qw(count1 count2 unknown10))}
#       = unpack (    's      s      s',    $buffer);
#
#      $in->read($buffer, 36 * $return{$s->('count1')});
#      $return{$s->('unknown11')} = [ map substr($buffer, $_*36, 36), 0..$return{$s->('count1')}-1 ];
#
#      $in->read($buffer, 8 * $return{$s->('count2')});
#      $return{$s->('unknown12')} = [ map substr($buffer, $_*8, 8), 0..$return{$s->('count2')}-1 ];
#
#      my $count = 0;
#      $count += unpack(L => $_) foreach @{$return{$s->('unknown12')}};
#
#      $in->read($buffer, $count);
#      $return{$s->('unknown13')} = $buffer;
#
#      if (exists $return{$s->('unknown09c')} and unpack l => $return{$s->('unknown09c')}) {
#        $in->read($buffer, unpack(l => $return{$s->('unknown09c')})*4);
#        $return{$s->('unknown14')} = $buffer;
#      } else {
#        $return{$s->('unknown14')} = '';
#      }
#    }
    if (defined $data{$s->('count1')}) {
      $out->print(pack (    's      s      s',
               @data{$s->(qw(count1 count2 unknown10))}));
      $out->print(@{$data{$s->('unknown11')}});
      $out->print(@{$data{$s->('unknown12')}});
      $out->print($data{$s->('unknown13')});
      $out->print($data{$s->('unknown14')});
    }

  }

  for my $matr_id (0..$data{num_materials}-1) {
    my $s = sub{wantarray?map"mtrl_${matr_id}_$_",@_:"mtrl_${matr_id}_$_[0]"};

#    $in->read($buffer, 4+2+10+4+36+12);
#    @return{$s->(qw(head id unknown01 name_length unknown02 unknown03))}
#     = unpack (    'a4   s  a10       l           a36       a12',    $buffer);

    $out->print(pack (    'a4   s  a10       l           a36       a12',
             @data{$s->(qw(head id unknown01 name_length unknown02 unknown03))}));

#    $return{$s->('head')} eq 'MTRL' or $die->("Invalid MTRL header (@{[$return{$s->('head')}]})"), return \%return;

#    $in->read($buffer, $return{$s->('name_length')});
#    $return{$s->('name')} = $buffer;

    $out->print($data{$s->('name')});

#    $in->read($buffer, 4);
#    $return{$s->('unknown04')} = $buffer;

    $out->print($data{$s->('unknown04')});


#    if ((unpack SSS => $return{$s->('unknown01')})[2] & 2) {
#      $in->read($buffer, 4);
#      $return{$s->('unknown05')} = $buffer;
#    }
    $out->print($data{$s->('unknown05')}) if defined $data{$s->('unknown05')};

#    if ( # erk... nasty code! [since I don't know how to detect whether there's a SFX]
#        $return{$s->('id')} == 591
#     or $return{$s->('id')} == 602
#     or $return{$s->('id')} == 803
#     or $return{$s->('id')} == 804
#     or $return{$s->('id')} == 876
#     or $return{$s->('id')} == 877
#     or ($return{$s->('name')} eq 'Naval N Kraken' and (
#           $return{$s->('id')} == 5 or
#           $return{$s->('id')} == 17 or
#           $return{$s->('id')} == 18 or
#           $return{$s->('id')} == 19 or
#           $return{$s->('id')} == 20 or
#           $return{$s->('id')} == 21 or
#           $return{$s->('id')} == 23 or
#           $return{$s->('id')} == 25 or
#           $return{$s->('id')} == 31))
#     or ($return{$s->('name')} eq 'Special N Battle Boar' and (
#           $return{$s->('id')} == 3 or
#           $return{$s->('id')} == 4))
#       ) {

    if (defined $data{$s->('num_sfx')}) {

#      $in->read($buffer, 1);
#      $return{$s->('num_sfx')} = unpack c => $buffer;

      $out->print(pack c => $data{$s->('num_sfx')});
  
      for my $sfx_id (0..$data{$s->('num_sfx')}-1) {
        my $s = sub{wantarray?map"mtrl_${matr_id}_sfx_${sfx_id}_$_",@_:"mtrl_${matr_id}_sfx_${sfx_id}_$_[0]"};
  
#        $in->read($buffer, 1);
#        $return{$s->('id')} = unpack c => $buffer;

        $out->print(pack c => $data{$s->('id')});
  
#        $in->read($buffer, 2);
#        $return{$s->('name_length')} = unpack s => $buffer;

        $out->print(pack s => $data{$s->('name_length')});
  
#        $in->read($buffer, $return{$s->('name_length')});
#        $return{$s->('name')} = $buffer;

        $out->print($data{$s->('name')});

#        die "Invalid SFX name @{[$return{$s->('name')}]}!" unless $return{$s->('name')} =~ /SFX/;
      }
    }
  }

#  if (not $in->eof) {
#    if (exists $return{mtrl_2_name} and $return{mtrl_2_name} eq 'Special N Raven') {
#      # "special n raven_birth.brg" has 4099 bytes of seemingly useless repeated data
#    } elsif (exists $return{mtrl_1_name} and $return{mtrl_1_name} eq 'world a breastplate map') {
#      # "world g weapons 4.brg" has 98 bytes of incorrect duplicate MTRL data
#    } else {
##      print "\n\n\n", $in->read($buffer, 1e6), ":\n", join('', unpack 'H*', $buffer);
#      die "Not EOF!";
#    }
#  }

#  return \%return;
  return;
}


sub threeds_to_brg__patch_data {
  my $file_data = shift;
  my %data = %{shift()};

  # From max's exported data:
  # Normals and verts need to be a,b,c -> -a,c,-b
  # TVerts are fine
  # Vertex mats need to be guessed (find faces using that vertex, etc)

  my @lines = split /\n/, $file_data;
  (shift @lines) =~ /^(\d+)~(\d+)~(\d+)f~$/ or die "Invalid header line!";
  my ($num_verts, $num_faces, $num_frames) = ($1, $2, $3);

  $num_frames = 1 if $num_frames == 0; # non-animated models

  (shift @lines) eq '{FACES}' or die "Invalid {FACES} line!";
  my @faces;
  my $faces = shift @lines;
  push @faces, [$2-1, $1-1, $3-1] while $faces =~ /\[(\d+),(\d+),(\d+)\]/g;
  @faces == $num_faces or die "Incorrect number of faces!";

  (shift @lines) eq '{TVERTS}' or die "Invalid {TVERTS} line!";
  my @tverts;
  my $tverts = shift @lines;
  push @tverts, [$1, $2] while $tverts =~ /\[([^,]+),([^,]+),([^\]]+)\]/g;
  @tverts == $num_verts or die "Incorrect number of texture vertices!";

  (shift @lines) eq '{VERTS}' or die "Invalid {VERTS} line!";
  my (@verts, @norms);
  my @verts_raw = split '~', shift @lines;
  @verts_raw == $num_verts * $num_frames or die "Incorrect number of vertices!";
  for my $f (0..$num_frames-1) {
    for (0..$num_verts-1) {
      (shift @verts_raw) =~ m!^([^ ]+) ([^ ]+) ([^/]+)/([^ ]+) ([^ ]+) ([^ ]+)$! or die "Invalid vertex data!";
      push @{$verts[$f]}, [-$1, $3, -$2];
      push @{$norms[$f]}, [-$4, $6, -$5];
    }
  }

  (shift @lines) eq '{FMATS}' or die "Invalid {FMATS} line!";
  my @face_mats = split '~', shift @lines;
  @face_mats == $num_faces or die "Incorrect number of face materials!";

  (shift @lines) eq '{MATS}' or die "Invalid {MATS} line!";
  my @mats = split '~', shift @lines;
  my @mat_ids;
  foreach (@mats) {
    /^(\d+)=.*mat_(\d{4})/ or die "Invalid material filename '$_'!";
    $mat_ids[$1] = $2;
  }

  # # #

  $num_frames == $data{num_meshes} or die "Cannot change the number of frames in the animation!";

  for my $mesh_id (0..$data{num_meshes}-1) {
    my $s = sub{wantarray?map"mesi_${mesh_id}_$_",@_:"mesi_${mesh_id}_$_[0]"};
    $data{$s->('num_vertices')} = $num_verts;
    $data{$s->('vertices')} = pack_float16(map @$_, @{$verts[$mesh_id]});
    $data{$s->('normals')}  = pack_float16(map @$_, @{$norms[$mesh_id]});
    if (length $data{$s->('texture_coords')}) {
      $data{$s->('texture_coords')} = pack_float16(map @$_, @tverts);
    }

    if (defined $data{$s->('unknown0a')}) {
      if (length $data{$s->('unknown0a')} < 4*$num_verts) {
        $data{$s->('unknown0a')} .= "\0\0\0\0" until length $data{$s->('unknown0a')} >= 4*$num_verts;
      } elsif (length $data{$s->('unknown0a')} > 4*$num_verts) {
        $data{$s->('unknown0a')} = substr($data{$s->('unknown0a')}, 0, 4*$num_verts);
      }
    }

    $data{$s->('num_faces')} = $num_faces;
  }

  $data{mesi_0_faces} = pack 's*', map @$_, @faces;
  $data{mesi_0_face_mats} = pack 's*', map $mat_ids[$_], @face_mats;

  # create vertex mats, based on the most common material of the faces which use that vertex
  my @vert_mats;
  for my $f (0..$#faces) {
    push @{$vert_mats[$faces[$f][$_]]}, $face_mats[$f] for 0..2;
  }
  $data{mesi_0_vertex_mats} = pack 's*', map $mat_ids[$_], map $vert_mats[$_]?commonest(@{$vert_mats[$_]}):$face_mats[0], 0..$num_verts-1;

  return \%data;
}


################################################
# MISCELLANEOUS STUFF:
################################################

sub commonest { my %f; ++$f{$_} for @_; return +(sort {$f{$b}<=>$f{$a}} keys %f)[0] }

sub utf16to8 { # this works backwards! (the 'wrong' byte order)
  my $return = '';
  $return .= $_<0x80 ? chr($_) :
            $_<0x800 ? chr(0xC0|($_>>6)).chr(0x80|($_&0x3F)) :
           chr(0xE0|($_>>12)).chr(0x80|(($_>>6)&0x3F)).chr(0x80|($_&0x3F))
            foreach unpack('S*', shift);
  $return;
}
sub utf8to16 { # this works backwards! (the 'wrong' byte order)
  my $return = '';

  # originally copied from Unicode::String
  my $bytes = $_[0];

  $bytes =~ s/^[\200-\277]+//;
  while (length $bytes) {
    if ($bytes =~ s/^([\000-\177]+)//) {
      $return .= pack("S*", unpack("C*", $1));
    } elsif ($bytes =~ s/^([\300-\337])([\200-\277])//) {
      $return .= pack("S", ((ord($1) & 0x1F) << 6) | (ord($2) & 0x3F));
    } elsif ($bytes =~ s/^([\340-\357])([\200-\277])([\200-\277])//) {
      $return .= pack("S", ((ord($1) & 0x0F) << 12) | ((ord($2) & 0x3F) <<  6) | (ord($3) & 0x3F));
    } else {
      die "Bad UTF-8 data!";
    }
  }
  $return;
}

sub xml_encode {
  return HTML::Entities::encode(@_);
}

sub xml_encode_lesser { # don't do any funny character translations
  my ($text) = @_;
  $text =~ s/&/&amp;/g;
  $text =~ s/"/&quot;/g;
  $text =~ s/</&lt;/g;
  $text =~ s/>/&gt;/g;
  $text;
}

sub xml_encode_muchlesser { # do even less translations
  my ($text) = @_;
  $text =~ s/&/&amp;/g;
  $text =~ s/</&lt;/g;
  $text =~ s/>/&gt;/g;
  $text;
}

sub unpack_float16 {
  my $data = $_[0];
  $data =~ s/(..)/\0\0$1/sg;
  return unpack 'f*' => $data;
}

sub pack_float16 {
  my $data = pack 'f*' => @_;
  $data =~ s/..(..)/$1/sg;
  return $data;
}

sub matlib_alpha { # doesn't return anything useful
  my ($matlib, $mat_id) = @_;
  my $fh = IO::File->new("< $matlib") or die "Cannot open materials.lib ('$matlib'): $!";

  my $buffer;

  $fh->seek(32 + 88+$mat_id*156, 0);
  $fh->read($buffer, 4);
  $fh->close;

  return unpack(L => $buffer) & 4 ? 0 : 1;
}

sub bar_extract_file {
  my ($barfile, $desired_filename) = @_;
  $desired_filename = lc $desired_filename;
  my $fh = IO::File->new("< $barfile") or die "Cannot open BAR '$barfile': $!";
  binmode $fh;

  my $buffer;

  $fh->seek(12, 0);

  $fh->read($buffer, 4);
  my $num_files = unpack l => $buffer;

  $fh->seek(4, 1);

  # Filetable position
  $fh->read($buffer, 4);
  my $ft_offset = unpack l => $buffer;

  $fh->seek($ft_offset, 0);

  # File offsets
  $fh->read($buffer, 4*$num_files);
  my @offsets = unpack 'l*' => $buffer;

  my $file_data;


  # Guess where the file is most likely to be alphabetically
  # (warning: attempted cleverness)
  my $f = int($num_files * (ord(substr($desired_filename, 0, 1))-97) / 25);

  $fh->seek($ft_offset + 4*$num_files + $offsets[$f], 0);
  $fh->read($buffer, 4+4+4+7+1+256);
  my ($offset, $length, $filename) = unpack 'llx4x7xZ*' => $buffer;

  my $f_dir = lc($filename) cmp $desired_filename;

  if ($f_dir == 0) {
    $fh->seek($offset, 0);
    $fh->read($file_data, $length);
    return $file_data;
  }

  my @f_ranges;
  if ($f_dir == 1) { # f is too high, so search down then up
    @f_ranges = (reverse(0..$f-1), 1+$f..$num_files-1);
  } else { # f is too low, so search up then down
    @f_ranges = (1+$f..$num_files-1, reverse(0..$f-1));
  }

  foreach (@f_ranges) {
    $fh->seek($ft_offset + 4*$num_files + $offsets[$_], 0);
    $fh->read($buffer, 4+4+4+7+1+256);
    my ($offset, $length, $filename) = unpack 'llx4x7xZ*' => $buffer;
    if (lc($filename) eq $desired_filename) {
      $fh->seek($offset, 0);
      $fh->read($file_data, $length);
      return $file_data;
    }
  }

  $fh->close;

  die qq{Cannot find '$desired_filename' inside BAR file '$barfile'!};
}

1;
